@@ -8,9 +8,14 @@
 #include <sys/types.h>
 
 #include "fio.h"
+#include "smalloc.h"
+#include "filehash.h"
 
 static int root_warn;
 
+/*
+ * Leaves f->fd open on success, caller must close
+ */
 static int extend_file(struct thread_data *td, struct fio_file *f)
 {
 	int r, new_layout = 0, unlink_file = 0, flags;
@@ -28,13 +33,15 @@ static int extend_file(struct thread_data *td, struct fio_file *f)
 	 * does that for operations involving reads, or for writes
 	 * where overwrite is set
 	 */
-	if (td_read(td) || (td_write(td) && td->o.overwrite))
+	if (td_read(td) || (td_write(td) && td->o.overwrite) ||
+	    (td_write(td) && td->io_ops->flags & FIO_NOEXTEND))
 		new_layout = 1;
 	if (td_write(td) && !td->o.overwrite)
 		unlink_file = 1;
 
-	if ((unlink_file || new_layout) && (f->flags & FIO_FILE_EXISTS)) {
-		if (unlink(f->file_name) < 0) {
+	if (unlink_file || new_layout) {
+		dprint(FD_FILE, "layout unlink %s\n", f->file_name);
+		if ((unlink(f->file_name) < 0) && (errno != ENOENT)) {
 			td_verror(td, errno, "unlink");
 			return 1;
 		}
@@ -61,12 +68,13 @@ static int extend_file(struct thread_data *td, struct fio_file *f)
 		goto err;
 	}
 
+#ifdef FIO_HAVE_FALLOCATE
 	dprint(FD_FILE, "fallocate file %s, size %llu\n", f->file_name,
 							f->real_file_size);
-	if (posix_fallocate(f->fd, 0, f->real_file_size) < 0) {
-		td_verror(td, errno, "posix_fallocate");
-		goto err;
-	}
+	r = posix_fallocate(f->fd, 0, f->real_file_size);
+	if (r < 0)
+		log_err("fio: posix_fallocate fails: %s\n", strerror(-r));
+#endif
 
 	b = malloc(td->o.max_bs[DDIR_WRITE]);
 	memset(b, 0, td->o.max_bs[DDIR_WRITE]);
@@ -92,15 +100,18 @@ static int extend_file(struct thread_data *td, struct fio_file *f)
 		}
 	}
 
-	if (td->terminate)
+	if (td->terminate) {
+		dprint(FD_FILE, "terminate unlink %s\n", f->file_name);
 		unlink(f->file_name);
-	else if (td->o.create_fsync)
-		fsync(f->fd);
+	} else if (td->o.create_fsync) {
+		if (fsync(f->fd) < 0) {
+			td_verror(td, errno, "fsync");
+			goto err;
+		}
+	}
 
 	free(b);
 done:
-	close(f->fd);
-	f->fd = -1;
 	return 0;
 err:
 	close(f->fd);
@@ -110,11 +121,13 @@ err:
 
 static unsigned long long get_rand_file_size(struct thread_data *td)
 {
-	unsigned long long ret;
+	unsigned long long ret, sized;
 	long r;
 
 	r = os_random_long(&td->file_size_state);
-	ret = td->o.file_size_low + (unsigned long long) ((double) (td->o.file_size_high - td->o.file_size_low) * (r / (RAND_MAX + 1.0)));
+	sized = td->o.file_size_high - td->o.file_size_low;
+	ret = (unsigned long long) ((double) sized * (r / (OS_RAND_MAX + 1.0)));
+	ret += td->o.file_size_low;
 	ret -= (ret % td->o.rw_min_bs);
 	return ret;
 }
@@ -165,7 +178,8 @@ static int get_file_size(struct thread_data *td, struct fio_file *f)
 		return ret;
 
 	if (f->file_offset > f->real_file_size) {
-		log_err("%s: offset extends end (%Lu > %Lu)\n", td->o.name, f->file_offset, f->real_file_size);
+		log_err("%s: offset extends end (%Lu > %Lu)\n", td->o.name,
+					f->file_offset, f->real_file_size);
 		return 1;
 	}
 
@@ -173,27 +187,33 @@ static int get_file_size(struct thread_data *td, struct fio_file *f)
 	return 0;
 }
 
-int file_invalidate_cache(struct thread_data *td, struct fio_file *f)
+static int __file_invalidate_cache(struct thread_data *td, struct fio_file *f,
+				   unsigned long long off,
+				   unsigned long long len)
 {
 	int ret = 0;
 
-	dprint(FD_IO, "invalidate cache (%d)\n", td->o.odirect);
+	if (len == -1ULL)
+		len = f->io_size;
+	if (off == -1ULL)
+		off = f->file_offset;
 
-	if (td->o.odirect)
-		return 0;
+	dprint(FD_IO, "invalidate cache %s: %llu/%llu\n", f->file_name, off,
+								len);
 
 	/*
 	 * FIXME: add blockdev flushing too
 	 */
 	if (f->mmap)
-		ret = madvise(f->mmap, f->io_size, MADV_DONTNEED);
-	else if (f->filetype == FIO_TYPE_FILE)
-		ret = fadvise(f->fd, f->file_offset, f->io_size, POSIX_FADV_DONTNEED);
-	else if (f->filetype == FIO_TYPE_BD) {
+		ret = madvise(f->mmap, len, MADV_DONTNEED);
+	else if (f->filetype == FIO_TYPE_FILE) {
+		ret = fadvise(f->fd, off, len, POSIX_FADV_DONTNEED);
+	} else if (f->filetype == FIO_TYPE_BD) {
 		ret = blockdev_invalidate_cache(f->fd);
 		if (ret < 0 && errno == EACCES && geteuid()) {
 			if (!root_warn) {
-				log_err("fio: only root may flush block devices. Cache flush bypassed!\n");
+				log_err("fio: only root may flush block "
+					"devices. Cache flush bypassed!\n");
 				root_warn = 1;
 			}
 			ret = 0;
@@ -204,22 +224,68 @@ int file_invalidate_cache(struct thread_data *td, struct fio_file *f)
 	if (ret < 0) {
 		td_verror(td, errno, "invalidate_cache");
 		return 1;
+	} else if (ret > 0) {
+		td_verror(td, ret, "invalidate_cache");
+		return 1;
 	}
 
 	return ret;
+
 }
 
-void generic_close_file(struct thread_data fio_unused *td, struct fio_file *f)
+int file_invalidate_cache(struct thread_data *td, struct fio_file *f)
 {
+	if (!(f->flags & FIO_FILE_OPEN))
+		return 0;
+
+	return __file_invalidate_cache(td, f, -1, -1);
+}
+
+int generic_close_file(struct thread_data fio_unused *td, struct fio_file *f)
+{
+	int ret = 0;
+
 	dprint(FD_FILE, "fd close %s\n", f->file_name);
-	close(f->fd);
+
+	remove_file_hash(f);
+
+	if (close(f->fd) < 0)
+		ret = errno;
+
 	f->fd = -1;
+	return ret;
+}
+
+static int file_lookup_open(struct fio_file *f, int flags)
+{
+	struct fio_file *__f;
+	int from_hash;
+
+	__f = lookup_file_hash(f->file_name);
+	if (__f) {
+		dprint(FD_FILE, "found file in hash %s\n", f->file_name);
+		/*
+		 * racy, need the __f->lock locked
+		 */
+		f->lock = __f->lock;
+		f->lock_owner = __f->lock_owner;
+		f->lock_batch = __f->lock_batch;
+		f->lock_ddir = __f->lock_ddir;
+		from_hash = 1;
+	} else {
+		dprint(FD_FILE, "file not found in hash %s\n", f->file_name);
+		from_hash = 0;
+	}
+
+	f->fd = open(f->file_name, flags, 0600);
+	return from_hash;
 }
 
 int generic_open_file(struct thread_data *td, struct fio_file *f)
 {
 	int is_std = 0;
 	int flags = 0;
+	int from_hash = 0;
 
 	dprint(FD_FILE, "fd open %s\n", f->file_name);
 
@@ -242,13 +308,12 @@ int generic_open_file(struct thread_data *td, struct fio_file *f)
 	if (td->o.sync_io)
 		flags |= O_SYNC;
 	if (f->filetype != FIO_TYPE_FILE)
-		flags |= O_NOATIME;
+		flags |= FIO_O_NOATIME;
 
 open_again:
 	if (td_write(td)) {
-		assert(!read_only);
-
-		flags |= O_RDWR;
+		if (!read_only)
+			flags |= O_RDWR;
 
 		if (f->filetype == FIO_TYPE_FILE)
 			flags |= O_CREAT;
@@ -256,7 +321,7 @@ open_again:
 		if (is_std)
 			f->fd = dup(STDOUT_FILENO);
 		else
-			f->fd = open(f->file_name, flags, 0600);
+			from_hash = file_lookup_open(f, flags);
 	} else {
 		if (f->filetype == FIO_TYPE_CHAR && !read_only)
 			flags |= O_RDWR;
@@ -266,15 +331,15 @@ open_again:
 		if (is_std)
 			f->fd = dup(STDIN_FILENO);
 		else
-			f->fd = open(f->file_name, flags);
+			from_hash = file_lookup_open(f, flags);
 	}
 
 	if (f->fd == -1) {
 		char buf[FIO_VERROR_SIZE];
 		int __e = errno;
 
-		if (errno == EPERM && (flags & O_NOATIME)) {
-			flags &= ~O_NOATIME;
+		if (errno == EPERM && (flags & FIO_O_NOATIME)) {
+			flags &= ~FIO_O_NOATIME;
 			goto open_again;
 		}
 
@@ -286,6 +351,18 @@ open_again:
 	if (get_file_size(td, f))
 		goto err;
 
+	if (!from_hash && f->fd != -1) {
+		if (add_file_hash(f)) {
+			int ret;
+
+			/*
+			 * OK to ignore, we haven't done anything with it
+			 */
+			ret = generic_close_file(td, f);
+			goto open_again;
+		}
+	}
+
 	return 0;
 err:
 	close(f->fd);
@@ -304,7 +381,8 @@ int open_files(struct thread_data *td)
 		err = td_io_open_file(td, f);
 		if (err) {
 			if (td->error == EMFILE) {
-				log_err("fio: limited open files to: %d\n", td->nr_open_files);
+				log_err("fio: limited open files to: %d\n",
+							td->nr_open_files);
 				td->o.open_files = td->nr_open_files;
 				err = 0;
 				clear_error(td);
@@ -335,6 +413,9 @@ static int get_file_sizes(struct thread_data *td)
 	int err = 0;
 
 	for_each_file(td, f, i) {
+		dprint(FD_FILE, "get file size for %p/%d/%p\n", f, i,
+								f->file_name);
+
 		if (td->io_ops->open_file(td, f)) {
 			if (td->error != ENOENT) {
 				log_err("%s\n", td->verror);
@@ -365,6 +446,9 @@ int setup_files(struct thread_data *td)
 
 	dprint(FD_FILE, "setup files\n");
 
+	if (td->o.read_iolog_file)
+		return 0;
+
 	/*
 	 * if ioengine defines a setup() method, it's responsible for
 	 * opening the files and setting f->real_file_size to indicate
@@ -421,16 +505,19 @@ int setup_files(struct thread_data *td)
 				f->io_size = f->real_file_size - f->file_offset;
 		} else if (f->real_file_size < td->o.file_size_low ||
 			   f->real_file_size > td->o.file_size_high) {
-			if (f->file_offset > td->o.file_size_low) 
+			if (f->file_offset > td->o.file_size_low)
 				goto err_offset;
 			/*
 			 * file size given. if it's fixed, use that. if it's a
 			 * range, generate a random size in-between.
 			 */
-			if (td->o.file_size_low == td->o.file_size_high)
-				f->io_size = td->o.file_size_low - f->file_offset;
-			else
-				f->io_size = get_rand_file_size(td) - f->file_offset;
+			if (td->o.file_size_low == td->o.file_size_high) {
+				f->io_size = td->o.file_size_low
+						- f->file_offset;
+			} else {
+				f->io_size = get_rand_file_size(td)
+						- f->file_offset;
+			}
 		} else
 			f->io_size = f->real_file_size - f->file_offset;
 
@@ -445,7 +532,7 @@ int setup_files(struct thread_data *td)
 			need_extend++;
 			extend_size += (f->io_size + f->file_offset);
 			f->flags |= FIO_FILE_EXTEND;
-		}	
+		}
 	}
 
 	if (!td->o.size || td->o.size > total_size)
@@ -456,19 +543,32 @@ int setup_files(struct thread_data *td)
 	 */
 	if (need_extend) {
 		temp_stall_ts = 1;
-		log_info("%s: Laying out IO file(s) (%u file(s) / %LuMiB)\n",
-			td->o.name, need_extend, extend_size >> 20);
+		if (!terse_output)
+			log_info("%s: Laying out IO file(s) (%u file(s) /"
+				 " %LuMiB)\n", td->o.name, need_extend,
+					extend_size >> 20);
 
 		for_each_file(td, f, i) {
+			unsigned long long old_len, extend_len;
+
 			if (!(f->flags & FIO_FILE_EXTEND))
 				continue;
 
 			assert(f->filetype == FIO_TYPE_FILE);
 			f->flags &= ~FIO_FILE_EXTEND;
+			old_len = f->real_file_size;
+			extend_len = f->io_size + f->file_offset - old_len;
 			f->real_file_size = (f->io_size + f->file_offset);
 			err = extend_file(td, f);
 			if (err)
 				break;
+			
+			err = __file_invalidate_cache(td, f, old_len,
+								extend_len);
+			close(f->fd);
+			f->fd = -1;
+			if (err)
+				break;
 		}
 		temp_stall_ts = 0;
 	}
@@ -497,19 +597,30 @@ int init_random_map(struct thread_data *td)
 	struct fio_file *f;
 	unsigned int i;
 
-	if (td->o.norandommap)
+	if (td->o.norandommap || !td_random(td))
 		return 0;
 
 	for_each_file(td, f, i) {
-		blocks = (f->real_file_size + td->o.rw_min_bs - 1) / (unsigned long long) td->o.rw_min_bs;
-		num_maps = (blocks + BLOCKS_PER_MAP-1)/ (unsigned long long) BLOCKS_PER_MAP;
-		f->file_map = malloc(num_maps * sizeof(long));
-		if (!f->file_map) {
-			log_err("fio: failed allocating random map. If running a large number of jobs, try the 'norandommap' option\n");
+		blocks = (f->real_file_size + td->o.rw_min_bs - 1) /
+				(unsigned long long) td->o.rw_min_bs;
+		num_maps = (blocks + BLOCKS_PER_MAP - 1) /
+				(unsigned long long) BLOCKS_PER_MAP;
+		f->file_map = smalloc(num_maps * sizeof(int));
+		if (f->file_map) {
+			f->num_maps = num_maps;
+			continue;
+		}
+		if (!td->o.softrandommap) {
+			log_err("fio: failed allocating random map. If running"
+				" a large number of jobs, try the 'norandommap'"
+				" option or set 'softrandommap'. Or give"
+				" a larger --alloc-size to fio.\n");
 			return 1;
 		}
-		f->num_maps = num_maps;
-		memset(f->file_map, 0, num_maps * sizeof(long));
+
+		log_info("fio: file %s failed allocating random map. Running "
+			 "job without.\n", f->file_name);
+		f->num_maps = 0;
 	}
 
 	return 0;
@@ -520,25 +631,39 @@ void close_files(struct thread_data *td)
 	struct fio_file *f;
 	unsigned int i;
 
+	for_each_file(td, f, i)
+		td_io_close_file(td, f);
+}
+
+void close_and_free_files(struct thread_data *td)
+{
+	struct fio_file *f;
+	unsigned int i;
+
 	dprint(FD_FILE, "close files\n");
 
 	for_each_file(td, f, i) {
-		if (td->o.unlink && f->filetype == FIO_TYPE_FILE)
+		if (td->o.unlink && f->filetype == FIO_TYPE_FILE) {
+			dprint(FD_FILE, "free unlink %s\n", f->file_name);
 			unlink(f->file_name);
+		}
 
 		td_io_close_file(td, f);
+		remove_file_hash(f);
 
-		free(f->file_name);
+		sfree(f->file_name);
 		f->file_name = NULL;
 
 		if (f->file_map) {
-			free(f->file_map);
+			sfree(f->file_map);
 			f->file_map = NULL;
 		}
+		sfree(f);
 	}
 
 	td->o.filename = NULL;
 	free(td->files);
+	td->files_index = 0;
 	td->files = NULL;
 	td->o.nr_files = 0;
 }
@@ -571,12 +696,19 @@ int add_file(struct thread_data *td, const char *fname)
 
 	dprint(FD_FILE, "add file %s\n", fname);
 
-	td->files = realloc(td->files, (cur_files + 1) * sizeof(*f));
-
-	f = &td->files[cur_files];
-	memset(f, 0, sizeof(*f));
+	f = smalloc(sizeof(*f));
+	if (!f) {
+		log_err("fio: smalloc OOM\n");
+		assert(0);
+	}
+		
 	f->fd = -1;
 
+	dprint(FD_FILE, "resize file array to %d files\n", cur_files + 1);
+
+	td->files = realloc(td->files, (cur_files + 1) * sizeof(f));
+	td->files[cur_files] = f;
+
 	/*
 	 * init function, io engine may not be loaded yet
 	 */
@@ -587,14 +719,35 @@ int add_file(struct thread_data *td, const char *fname)
 		len = sprintf(file_name, "%s/", td->o.directory);
 
 	sprintf(file_name + len, "%s", fname);
-	f->file_name = strdup(file_name);
-
+	f->file_name = smalloc_strdup(file_name);
+	if (!f->file_name) {
+		log_err("fio: smalloc OOM\n");
+		assert(0);
+	}
+	
 	get_file_type(f);
 
+	switch (td->o.file_lock_mode) {
+	case FILE_LOCK_NONE:
+		break;
+	case FILE_LOCK_READWRITE:
+		f->lock = fio_mutex_rw_init();
+		break;
+	case FILE_LOCK_EXCLUSIVE:
+		f->lock = fio_mutex_init(1);
+		break;
+	default:
+		log_err("fio: unknown lock mode: %d\n", td->o.file_lock_mode);
+		assert(0);
+	}
+
 	td->files_index++;
 	if (f->filetype == FIO_TYPE_FILE)
 		td->nr_normal_files++;
 
+	dprint(FD_FILE, "file %p \"%s\" added at %d\n", f, f->file_name,
+							cur_files);
+
 	return cur_files;
 }
 
@@ -605,25 +758,89 @@ void get_file(struct fio_file *f)
 	f->references++;
 }
 
-void put_file(struct thread_data *td, struct fio_file *f)
+int put_file(struct thread_data *td, struct fio_file *f)
 {
+	int f_ret = 0, ret = 0;
+
 	dprint(FD_FILE, "put file %s, ref=%d\n", f->file_name, f->references);
 
 	if (!(f->flags & FIO_FILE_OPEN))
-		return;
+		return 0;
 
 	assert(f->references);
 	if (--f->references)
-		return;
+		return 0;
 
 	if (should_fsync(td) && td->o.fsync_on_close)
-		fsync(f->fd);
+		f_ret = fsync(f->fd);
 
 	if (td->io_ops->close_file)
-		td->io_ops->close_file(td, f);
+		ret = td->io_ops->close_file(td, f);
+
+	if (!ret)
+		ret = f_ret;
 
 	td->nr_open_files--;
 	f->flags &= ~FIO_FILE_OPEN;
+	return ret;
+}
+
+void lock_file(struct thread_data *td, struct fio_file *f, enum fio_ddir ddir)
+{
+	if (!f->lock || td->o.file_lock_mode == FILE_LOCK_NONE)
+		return;
+
+	if (f->lock_owner == td && f->lock_batch--)
+		return;
+
+	if (td->o.file_lock_mode == FILE_LOCK_READWRITE) {
+		if (ddir == DDIR_READ)
+			fio_mutex_down_read(f->lock);
+		else
+			fio_mutex_down_write(f->lock);
+	} else if (td->o.file_lock_mode == FILE_LOCK_EXCLUSIVE)
+		fio_mutex_down(f->lock);
+
+	f->lock_owner = td;
+	f->lock_batch = td->o.lockfile_batch;
+	f->lock_ddir = ddir;
+}
+
+void unlock_file(struct thread_data *td, struct fio_file *f)
+{
+	if (!f->lock || td->o.file_lock_mode == FILE_LOCK_NONE)
+		return;
+	if (f->lock_batch)
+		return;
+
+	if (td->o.file_lock_mode == FILE_LOCK_READWRITE) {
+		const int is_read = f->lock_ddir == DDIR_READ;
+		int val = fio_mutex_getval(f->lock);
+
+		if ((is_read && val == 1) || (!is_read && val == -1))
+			f->lock_owner = NULL;
+
+		if (is_read)
+			fio_mutex_up_read(f->lock);
+		else
+			fio_mutex_up_write(f->lock);
+	} else if (td->o.file_lock_mode == FILE_LOCK_EXCLUSIVE) {
+		int val = fio_mutex_getval(f->lock);
+
+		if (val == 0)
+			f->lock_owner = NULL;
+
+		fio_mutex_up(f->lock);
+	}
+}
+
+void unlock_file_all(struct thread_data *td, struct fio_file *f)
+{
+	if (f->lock_owner != td)
+		return;
+
+	f->lock_batch = 0;
+	unlock_file(td, f);
 }
 
 static int recurse_dir(struct thread_data *td, const char *dirname)
@@ -665,7 +882,8 @@ static int recurse_dir(struct thread_data *td, const char *dirname)
 		if (!S_ISDIR(sb.st_mode))
 			continue;
 
-		if ((ret = recurse_dir(td, full_path)) != 0)
+		ret = recurse_dir(td, full_path);
+		if (ret)
 			break;
 	}
 
@@ -687,18 +905,34 @@ void dup_files(struct thread_data *td, struct thread_data *org)
 {
 	struct fio_file *f;
 	unsigned int i;
-	size_t bytes;
+
+	dprint(FD_FILE, "dup files: %d\n", org->files_index);
 
 	if (!org->files)
 		return;
 
-	bytes = org->files_index * sizeof(*f);
-	td->files = malloc(bytes);
-	memcpy(td->files, org->files, bytes);
+	td->files = malloc(org->files_index * sizeof(f));
 
-	for_each_file(td, f, i) {
-		if (f->file_name)
-			f->file_name = strdup(f->file_name);
+	for_each_file(org, f, i) {
+		struct fio_file *__f;
+
+		__f = smalloc(sizeof(*__f));
+		if (!__f) {
+			log_err("fio: smalloc OOM\n");
+			assert(0);
+		}
+	
+		if (f->file_name) {
+			__f->file_name = smalloc_strdup(f->file_name);
+			if (!__f->file_name) {
+				log_err("fio: smalloc OOM\n");
+				assert(0);
+			}
+	
+			__f->filetype = f->filetype;
+		}
+
+		td->files[i] = __f;
 	}
 }
 
