@@ -36,6 +36,7 @@
 
 #include "fio.h"
 #include "hash.h"
+#include "smalloc.h"
 
 unsigned long page_mask;
 unsigned long page_size;
@@ -48,20 +49,25 @@ int nr_process = 0;
 int nr_thread = 0;
 int shm_id = 0;
 int temp_stall_ts;
+unsigned long done_secs = 0;
 
-static struct fio_sem *startup_sem;
+static struct fio_mutex *startup_mutex;
 static volatile int fio_abort;
 static int exit_value;
+static struct itimerval itimer;
 
 struct io_log *agg_io_log[2];
 
 #define TERMINATE_ALL		(-1)
 #define JOB_START_TIMEOUT	(5 * 1000)
 
-static inline void td_set_runstate(struct thread_data *td, int runstate)
+void td_set_runstate(struct thread_data *td, int runstate)
 {
-	dprint(FD_PROCESS, "pid=%d: runstate %d -> %d\n", td->pid, td->runstate,
-								runstate);
+	if (td->runstate == runstate)
+		return;
+
+	dprint(FD_PROCESS, "pid=%d: runstate %d -> %d\n", (int) td->pid,
+						td->runstate, runstate);
 	td->runstate = runstate;
 }
 
@@ -70,10 +76,12 @@ static void terminate_threads(int group_id)
 	struct thread_data *td;
 	int i;
 
+	dprint(FD_PROCESS, "terminate group_id=%d\n", group_id);
+
 	for_each_td(td, i) {
 		if (group_id == TERMINATE_ALL || groupid == td->groupid) {
-			dprint(FD_PROCESS, "setting terminate on %d\n",td->pid);
-
+			dprint(FD_PROCESS, "setting terminate on %s/%d\n",
+						td->o.name, (int) td->pid);
 			td->terminate = 1;
 			td->o.start_delay = 0;
 
@@ -82,28 +90,72 @@ static void terminate_threads(int group_id)
 			 */
 			if (td->runstate < TD_RUNNING)
 				kill(td->pid, SIGQUIT);
-			else if (td->io_ops->flags & FIO_SIGQUIT)
-				kill(td->pid, SIGQUIT);
+			else {
+				struct ioengine_ops *ops = td->io_ops;
+
+				if (ops && (ops->flags & FIO_SIGQUIT))
+					kill(td->pid, SIGQUIT);
+			}
 		}
 	}
 }
 
-static void sig_handler(int sig)
+static void status_timer_arm(void)
 {
-	switch (sig) {
-		case SIGALRM:
-			update_io_ticks();
-			disk_util_timer_arm();
-			print_thread_status();
-			break;
-		default:
-			printf("\nfio: terminating on signal %d\n", sig);
-			fflush(stdout);
-			terminate_threads(TERMINATE_ALL);
-			break;
+	itimer.it_value.tv_sec = 0;
+	itimer.it_value.tv_usec = DISK_UTIL_MSEC * 1000;
+	setitimer(ITIMER_REAL, &itimer, NULL);
+}
+
+static void sig_alrm(int fio_unused sig)
+{
+	if (threads) {
+		update_io_ticks();
+		print_thread_status();
+		status_timer_arm();
 	}
 }
 
+static void sig_int(int sig)
+{
+	if (threads) {
+		printf("\nfio: terminating on signal %d\n", sig);
+		fflush(stdout);
+		terminate_threads(TERMINATE_ALL);
+	}
+}
+
+static void sig_ill(int fio_unused sig)
+{
+	if (!threads)
+		return;
+
+	log_err("fio: illegal instruction. your cpu does not support "
+		"the sse4.2 instruction for crc32c\n");
+	terminate_threads(TERMINATE_ALL);
+	exit(4);
+}
+
+static void set_sig_handlers(void)
+{
+	struct sigaction act;
+
+	memset(&act, 0, sizeof(act));
+	act.sa_handler = sig_alrm;
+	act.sa_flags = SA_RESTART;
+	sigaction(SIGALRM, &act, NULL);
+
+	memset(&act, 0, sizeof(act));
+	act.sa_handler = sig_int;
+	act.sa_flags = SA_RESTART;
+	sigaction(SIGINT, &act, NULL);
+
+	memset(&act, 0, sizeof(act));
+	act.sa_handler = sig_ill;
+	act.sa_flags = SA_RESTART;
+	sigaction(SIGILL, &act, NULL);
+}
+
 /*
  * Check if we are above the minimum rate given.
  */
@@ -148,12 +200,16 @@ static int check_min_rate(struct thread_data *td, struct timeval *now)
 			 * check bandwidth specified rate
 			 */
 			if (bytes < td->rate_bytes) {
-				log_err("%s: min rate %u not met\n", td->o.name, td->o.ratemin);
+				log_err("%s: min rate %u not met\n", td->o.name,
+								td->o.ratemin);
 				return 1;
 			} else {
 				rate = (bytes - td->rate_bytes) / spent;
-				if (rate < td->o.ratemin || bytes < td->rate_bytes) {
-					log_err("%s: min rate %u not met, got %luKiB/sec\n", td->o.name, td->o.ratemin, rate);
+				if (rate < td->o.ratemin ||
+				    bytes < td->rate_bytes) {
+					log_err("%s: min rate %u not met, got"
+						" %luKiB/sec\n", td->o.name,
+							td->o.ratemin, rate);
 					return 1;
 				}
 			}
@@ -162,12 +218,17 @@ static int check_min_rate(struct thread_data *td, struct timeval *now)
 			 * checks iops specified rate
 			 */
 			if (iops < td->o.rate_iops) {
-				log_err("%s: min iops rate %u not met\n", td->o.name, td->o.rate_iops);
+				log_err("%s: min iops rate %u not met\n",
+						td->o.name, td->o.rate_iops);
 				return 1;
 			} else {
 				rate = (iops - td->rate_blocks) / spent;
-				if (rate < td->o.rate_iops_min || iops < td->rate_blocks) {
-					log_err("%s: min iops rate %u not met, got %lu\n", td->o.name, td->o.rate_iops_min, rate);
+				if (rate < td->o.rate_iops_min ||
+				    iops < td->rate_blocks) {
+					log_err("%s: min iops rate %u not met,"
+						" got %lu\n", td->o.name,
+							td->o.rate_iops_min,
+							rate);
 				}
 			}
 		}
@@ -195,7 +256,7 @@ static inline int runtime_exceeded(struct thread_data *td, struct timeval *t)
  */
 static void cleanup_pending_aio(struct thread_data *td)
 {
-	struct list_head *entry, *n;
+	struct flist_head *entry, *n;
 	struct io_u *io_u;
 	int r;
 
@@ -210,8 +271,8 @@ static void cleanup_pending_aio(struct thread_data *td)
 	 * now cancel remaining active events
 	 */
 	if (td->io_ops->cancel) {
-		list_for_each_safe(entry, n, &td->io_u_busylist) {
-			io_u = list_entry(entry, struct io_u, list);
+		flist_for_each_safe(entry, n, &td->io_u_busylist) {
+			io_u = flist_entry(entry, struct io_u, list);
 
 			/*
 			 * if the io_u isn't in flight, then that generally
@@ -310,7 +371,7 @@ static void do_verify(struct thread_data *td)
 
 	io_u = NULL;
 	while (!td->terminate) {
-		int ret2;
+		int ret2, full;
 
 		io_u = __get_io_u(td);
 		if (!io_u)
@@ -347,7 +408,7 @@ static void do_verify(struct thread_data *td)
 				 * zero read, fail
 				 */
 				if (!bytes) {
-					td_verror(td, ENODATA, "full resid");
+					td_verror(td, EIO, "full resid");
 					put_io_u(td, io_u);
 					break;
 				}
@@ -355,7 +416,6 @@ static void do_verify(struct thread_data *td)
 				io_u->xfer_buflen = io_u->resid;
 				io_u->xfer_buf += bytes;
 				io_u->offset += bytes;
-				f->last_completed_pos = io_u->offset;
 
 				td->ts.short_io_u[io_u->ddir]++;
 
@@ -391,19 +451,25 @@ sync_done:
 		 * if we can queue more, do so. but check if there are
 		 * completed io_u's first.
 		 */
-		min_events = 0;
-		if (queue_full(td) || ret == FIO_Q_BUSY) {
-			min_events = 1;
+		full = queue_full(td) || ret == FIO_Q_BUSY;
+		if (full || !td->o.iodepth_batch_complete) {
+			min_events = td->o.iodepth_batch_complete;
+			if (full && !min_events)
+				min_events = 1;
 
-			if (td->cur_depth > td->o.iodepth_low)
-				min_events = td->cur_depth - td->o.iodepth_low;
+			do {
+				/*
+				 * Reap required number of io units, if any,
+				 * and do the verification on them through
+				 * the callback handler
+				 */
+				if (io_u_queued_complete(td, min_events) < 0) {
+					ret = -1;
+					break;
+				}
+			} while (full && (td->cur_depth > td->o.iodepth_low));
 		}
-
-		/*
-		 * Reap required number of io units, if any, and do the
-		 * verification on them through the callback handler
-		 */
-		if (io_u_queued_complete(td, min_events) < 0)
+		if (ret < 0)
 			break;
 	}
 
@@ -429,14 +495,17 @@ static void do_io(struct thread_data *td)
 	unsigned int i;
 	int ret = 0;
 
-	td_set_runstate(td, TD_RUNNING);
+	if (in_ramp_time(td))
+		td_set_runstate(td, TD_RAMP);
+	else
+		td_set_runstate(td, TD_RUNNING);
 
 	while ((td->this_io_bytes[0] + td->this_io_bytes[1]) < td->o.size) {
 		struct timeval comp_time;
 		long bytes_done = 0;
 		int min_evts = 0;
 		struct io_u *io_u;
-		int ret2;
+		int ret2, full;
 
 		if (td->terminate)
 			break;
@@ -457,8 +526,13 @@ static void do_io(struct thread_data *td)
 		 * Add verification end_io handler, if asked to verify
 		 * a previously written file.
 		 */
-		if (td->o.verify != VERIFY_NONE)
+		if (td->o.verify != VERIFY_NONE && io_u->ddir == DDIR_READ) {
 			io_u->end_io = verify_io_u;
+			td_set_runstate(td, TD_VERIFYING);
+		} else if (in_ramp_time(td))
+			td_set_runstate(td, TD_RAMP);
+		else
+			td_set_runstate(td, TD_RUNNING);
 
 		ret = td_io_queue(td, io_u);
 		switch (ret) {
@@ -473,7 +547,7 @@ static void do_io(struct thread_data *td)
 				 * zero read, fail
 				 */
 				if (!bytes) {
-					td_verror(td, ENODATA, "full resid");
+					td_verror(td, EIO, "full resid");
 					put_io_u(td, io_u);
 					break;
 				}
@@ -481,7 +555,6 @@ static void do_io(struct thread_data *td)
 				io_u->xfer_buflen = io_u->resid;
 				io_u->xfer_buf += bytes;
 				io_u->offset += bytes;
-				f->last_completed_pos = io_u->offset;
 
 				td->ts.short_io_u[io_u->ddir]++;
 
@@ -524,21 +597,25 @@ sync_done:
 		/*
 		 * See if we need to complete some commands
 		 */
-		if (ret == FIO_Q_QUEUED || ret == FIO_Q_BUSY) {
-			min_evts = 0;
-			if (queue_full(td) || ret == FIO_Q_BUSY) {
+		full = queue_full(td) || ret == FIO_Q_BUSY;
+		if (full || !td->o.iodepth_batch_complete) {
+			min_evts = td->o.iodepth_batch_complete;
+			if (full && !min_evts)
 				min_evts = 1;
 
-				if (td->cur_depth > td->o.iodepth_low)
-					min_evts = td->cur_depth - td->o.iodepth_low;
-			}
-
 			fio_gettime(&comp_time, NULL);
-			bytes_done = io_u_queued_complete(td, min_evts);
-			if (bytes_done < 0)
-				break;
+
+			do {
+				ret = io_u_queued_complete(td, min_evts);
+				if (ret <= 0)
+					break;
+
+				bytes_done += ret;
+			} while (full && (td->cur_depth > td->o.iodepth_low));
 		}
 
+		if (ret < 0)
+			break;
 		if (!bytes_done)
 			continue;
 
@@ -547,15 +624,17 @@ sync_done:
 		 * of completions except the very first one which may look
 		 * a little bursty
 		 */
-		usec = utime_since(&s, &comp_time);
+		if (!in_ramp_time(td)) {
+			usec = utime_since(&s, &comp_time);
 
-		rate_throttle(td, usec, bytes_done);
+			rate_throttle(td, usec, bytes_done);
 
-		if (check_min_rate(td, &comp_time)) {
-			if (exitall_on_terminate)
-				terminate_threads(td->groupid);
-			td_verror(td, ENODATA, "check_min_rate");
-			break;
+			if (check_min_rate(td, &comp_time)) {
+				if (exitall_on_terminate)
+					terminate_threads(td->groupid);
+				td_verror(td, EIO, "check_min_rate");
+				break;
+			}
 		}
 
 		if (td->o.thinktime) {
@@ -607,35 +686,19 @@ sync_done:
 
 static void cleanup_io_u(struct thread_data *td)
 {
-	struct list_head *entry, *n;
+	struct flist_head *entry, *n;
 	struct io_u *io_u;
 
-	list_for_each_safe(entry, n, &td->io_u_freelist) {
-		io_u = list_entry(entry, struct io_u, list);
+	flist_for_each_safe(entry, n, &td->io_u_freelist) {
+		io_u = flist_entry(entry, struct io_u, list);
 
-		list_del(&io_u->list);
+		flist_del(&io_u->list);
 		free(io_u);
 	}
 
 	free_io_mem(td);
 }
 
-/*
- * "randomly" fill the buffer contents
- */
-static void fill_io_buf(struct thread_data *td, struct io_u *io_u, int max_bs)
-{
-	long *ptr = io_u->buf;
-
-	if (!td->o.zero_buffers) {
-		while ((void *) ptr - io_u->buf < max_bs) {
-			*ptr = rand() * GOLDEN_RATIO_PRIME;
-			ptr++;
-		}
-	} else
-		memset(ptr, 0, max_bs);
-}
-
 static int init_io_u(struct thread_data *td)
 {
 	struct io_u *io_u;
@@ -645,10 +708,15 @@ static int init_io_u(struct thread_data *td)
 
 	max_units = td->o.iodepth;
 	max_bs = max(td->o.max_bs[DDIR_READ], td->o.max_bs[DDIR_WRITE]);
-	td->orig_buffer_size = (unsigned long long) max_bs * (unsigned long long) max_units;
+	td->orig_buffer_size = (unsigned long long) max_bs
+					* (unsigned long long) max_units;
 
-	if (td->o.mem_type == MEM_SHMHUGE || td->o.mem_type == MEM_MMAPHUGE)
-		td->orig_buffer_size = (td->orig_buffer_size + td->o.hugepage_size - 1) & ~(td->o.hugepage_size - 1);
+	if (td->o.mem_type == MEM_SHMHUGE || td->o.mem_type == MEM_MMAPHUGE) {
+		unsigned long bs;
+
+		bs = td->orig_buffer_size + td->o.hugepage_size - 1;
+		td->orig_buffer_size = bs & ~(td->o.hugepage_size - 1);
+	}
 
 	if (td->orig_buffer_size != (size_t) td->orig_buffer_size) {
 		log_err("fio: IO memory too large. Reduce max_bs or iodepth\n");
@@ -668,18 +736,18 @@ static int init_io_u(struct thread_data *td)
 			return 1;
 		io_u = malloc(sizeof(*io_u));
 		memset(io_u, 0, sizeof(*io_u));
-		INIT_LIST_HEAD(&io_u->list);
+		INIT_FLIST_HEAD(&io_u->list);
 
 		if (!(td->io_ops->flags & FIO_NOIO)) {
 			io_u->buf = p + max_bs * i;
 
-			if (td_write(td))
-				fill_io_buf(td, io_u, max_bs);
+			if (td_write(td) && !td->o.refill_buffers)
+				io_u_fill_buffer(td, io_u, max_bs);
 		}
 
 		io_u->index = i;
 		io_u->flags = IO_U_F_FREE;
-		list_add(&io_u->list, &td->io_u_freelist);
+		flist_add(&io_u->list, &td->io_u_freelist);
 	}
 
 	io_u_init_timeout();
@@ -701,7 +769,8 @@ static int switch_ioscheduler(struct thread_data *td)
 	f = fopen(tmp, "r+");
 	if (!f) {
 		if (errno == ENOENT) {
-			log_err("fio: os or kernel doesn't support IO scheduler switching\n");
+			log_err("fio: os or kernel doesn't support IO scheduler"
+				" switching\n");
 			return 0;
 		}
 		td_verror(td, errno, "fopen iosched");
@@ -755,19 +824,16 @@ static int keep_running(struct thread_data *td)
 		return 1;
 	}
 
-	io_done = td->io_bytes[DDIR_READ] + td->io_bytes[DDIR_WRITE] + td->io_skip_bytes;
+	io_done = td->io_bytes[DDIR_READ] + td->io_bytes[DDIR_WRITE]
+			+ td->io_skip_bytes;
 	if (io_done < td->o.size)
 		return 1;
 
 	return 0;
 }
 
-static int clear_io_state(struct thread_data *td)
+static void reset_io_counters(struct thread_data *td)
 {
-	struct fio_file *f;
-	unsigned int i;
-	int ret;
-
 	td->ts.stat_io_bytes[0] = td->ts.stat_io_bytes[1] = 0;
 	td->this_io_bytes[0] = td->this_io_bytes[1] = 0;
 	td->zone_bytes = 0;
@@ -782,9 +848,36 @@ static int clear_io_state(struct thread_data *td)
 	 */
 	if (td->o.time_based || td->o.loops)
 		td->nr_done_files = 0;
+}
 
-	for_each_file(td, f, i)
-		td_io_close_file(td, f);
+void reset_all_stats(struct thread_data *td)
+{
+	struct timeval tv;
+	int i;
+
+	reset_io_counters(td);
+
+	for (i = 0; i < 2; i++) {
+		td->io_bytes[i] = 0;
+		td->io_blocks[i] = 0;
+		td->io_issues[i] = 0;
+		td->ts.total_io_u[i] = 0;
+	}
+	
+	fio_gettime(&tv, NULL);
+	memcpy(&td->epoch, &tv, sizeof(tv));
+	memcpy(&td->start, &tv, sizeof(tv));
+}
+
+static int clear_io_state(struct thread_data *td)
+{
+	struct fio_file *f;
+	unsigned int i;
+	int ret;
+
+	reset_io_counters(td);
+
+	close_files(td);
 
 	ret = 0;
 	for_each_file(td, f, i) {
@@ -812,24 +905,24 @@ static void *thread_main(void *data)
 
 	td->pid = getpid();
 
-	dprint(FD_PROCESS, "jobs pid=%d started\n", td->pid);
+	dprint(FD_PROCESS, "jobs pid=%d started\n", (int) td->pid);
 
-	INIT_LIST_HEAD(&td->io_u_freelist);
-	INIT_LIST_HEAD(&td->io_u_busylist);
-	INIT_LIST_HEAD(&td->io_u_requeues);
-	INIT_LIST_HEAD(&td->io_log_list);
-	INIT_LIST_HEAD(&td->io_hist_list);
+	INIT_FLIST_HEAD(&td->io_u_freelist);
+	INIT_FLIST_HEAD(&td->io_u_busylist);
+	INIT_FLIST_HEAD(&td->io_u_requeues);
+	INIT_FLIST_HEAD(&td->io_log_list);
+	INIT_FLIST_HEAD(&td->io_hist_list);
 	td->io_hist_tree = RB_ROOT;
 
 	td_set_runstate(td, TD_INITIALIZED);
-	fio_sem_up(startup_sem);
-	fio_sem_down(td->mutex);
+	fio_mutex_up(startup_mutex);
+	fio_mutex_down(td->mutex);
 
 	/*
-	 * the ->mutex semaphore is now no longer used, close it to avoid
+	 * the ->mutex mutex is now no longer used, close it to avoid
 	 * eating a file descriptor
 	 */
-	fio_sem_remove(td->mutex);
+	fio_mutex_remove(td->mutex);
 
 	/*
 	 * May alter parameters that init_io_u() will use, so we need to
@@ -889,7 +982,8 @@ static void *thread_main(void *data)
 		memcpy(&td->ts.stat_sample_time, &td->start, sizeof(td->start));
 
 		if (td->o.ratemin)
-			memcpy(&td->lastrate, &td->ts.stat_sample_time, sizeof(td->lastrate));
+			memcpy(&td->lastrate, &td->ts.stat_sample_time,
+							sizeof(td->lastrate));
 
 		if (clear_state && clear_io_state(td))
 			break;
@@ -902,7 +996,8 @@ static void *thread_main(void *data)
 
 		if (td_read(td) && td->io_bytes[DDIR_READ]) {
 			if (td->rw_end_set[DDIR_READ])
-				elapsed = utime_since(&td->start, &td->rw_end[DDIR_READ]);
+				elapsed = utime_since(&td->start,
+						      &td->rw_end[DDIR_READ]);
 			else
 				elapsed = utime_since_now(&td->start);
 
@@ -910,13 +1005,14 @@ static void *thread_main(void *data)
 		}
 		if (td_write(td) && td->io_bytes[DDIR_WRITE]) {
 			if (td->rw_end_set[DDIR_WRITE])
-				elapsed = utime_since(&td->start, &td->rw_end[DDIR_WRITE]);
+				elapsed = utime_since(&td->start,
+						      &td->rw_end[DDIR_WRITE]);
 			else
 				elapsed = utime_since_now(&td->start);
 
 			runtime[DDIR_WRITE] += elapsed;
 		}
-		
+
 		if (td->error || td->terminate)
 			break;
 
@@ -961,8 +1057,9 @@ static void *thread_main(void *data)
 
 err:
 	if (td->error)
-		printf("fio: pid=%d, err=%d/%s\n", td->pid, td->error, td->verror);
-	close_files(td);
+		printf("fio: pid=%d, err=%d/%s\n", (int) td->pid, td->error,
+							td->verror);
+	close_and_free_files(td);
 	close_ioengine(td);
 	cleanup_io_u(td);
 
@@ -1024,7 +1121,11 @@ static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
 		else
 			realthreads++;
 
-		if (!td->pid || td->runstate == TD_REAPED)
+		if (!td->pid) {
+			pending++;
+			continue;
+		}
+		if (td->runstate == TD_REAPED)
 			continue;
 		if (td->o.use_thread) {
 			if (td->runstate == TD_EXITED) {
@@ -1044,7 +1145,8 @@ static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
 		ret = waitpid(td->pid, &status, flags);
 		if (ret < 0) {
 			if (errno == ECHILD) {
-				log_err("fio: pid=%d disappeared %d\n", td->pid, td->runstate);
+				log_err("fio: pid=%d disappeared %d\n",
+						(int) td->pid, td->runstate);
 				td_set_runstate(td, TD_REAPED);
 				goto reaped;
 			}
@@ -1054,7 +1156,8 @@ static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
 				int sig = WTERMSIG(status);
 
 				if (sig != SIGQUIT)
-					log_err("fio: pid=%d, got signal=%d\n", td->pid, sig);
+					log_err("fio: pid=%d, got signal=%d\n",
+							(int) td->pid, sig);
 				td_set_runstate(td, TD_REAPED);
 				goto reaped;
 			}
@@ -1073,23 +1176,16 @@ static void reap_threads(int *nr_running, int *t_rate, int *m_rate)
 		pending++;
 		continue;
 reaped:
-		if (td->o.use_thread) {
-			long ret;
-
-			dprint(FD_PROCESS, "joining tread %d\n", td->pid);
-			if (pthread_join(td->thread, (void *) &ret)) {
-				dprint(FD_PROCESS, "join failed %ld\n", ret);
-				perror("pthread_join");
-			}
-		}
-
 		(*nr_running)--;
 		(*m_rate) -= td->o.ratemin;
 		(*t_rate) -= td->o.rate;
-		pending--;
+		if (!td->pid)
+			pending--;
 
 		if (td->error)
 			exit_value++;
+
+		done_secs += mtime_since_now(&td->epoch) / 1000;
 	}
 
 	if (*nr_running == cputhreads && !pending && realthreads)
@@ -1111,18 +1207,19 @@ static void run_threads(void)
 	if (!terse_output) {
 		printf("Starting ");
 		if (nr_thread)
-			printf("%d thread%s", nr_thread, nr_thread > 1 ? "s" : "");
+			printf("%d thread%s", nr_thread,
+						nr_thread > 1 ? "s" : "");
 		if (nr_process) {
 			if (nr_thread)
 				printf(" and ");
-			printf("%d process%s", nr_process, nr_process > 1 ? "es" : "");
+			printf("%d process%s", nr_process,
+						nr_process > 1 ? "es" : "");
 		}
 		printf("\n");
 		fflush(stdout);
 	}
 
-	signal(SIGINT, sig_handler);
-	signal(SIGALRM, sig_handler);
+	set_sig_handlers();
 
 	todo = thread_number;
 	nr_running = 0;
@@ -1145,9 +1242,21 @@ static void run_threads(void)
 		if (setup_files(td)) {
 			exit_value++;
 			if (td->error)
-				log_err("fio: pid=%d, err=%d/%s\n", td->pid, td->error, td->verror);
+				log_err("fio: pid=%d, err=%d/%s\n",
+					(int) td->pid, td->error, td->verror);
 			td_set_runstate(td, TD_REAPED);
 			todo--;
+		} else {
+			struct fio_file *f;
+			unsigned int i;
+
+			/*
+			 * for sharing to work, each job must always open
+			 * its own files. so close them, if we opened them
+			 * for creation
+			 */
+			for_each_file(td, f, i)
+				td_io_close_file(td, f);
 		}
 
 		init_disk_util(td);
@@ -1183,8 +1292,11 @@ static void run_threads(void)
 					continue;
 			}
 
-			if (td->o.stonewall && (nr_started || nr_running))
+			if (td->o.stonewall && (nr_started || nr_running)) {
+				dprint(FD_PROCESS, "%s: stonewall wait\n",
+							td->o.name);
 				break;
+			}
 
 			/*
 			 * Set state to created. Thread will transition
@@ -1196,20 +1308,26 @@ static void run_threads(void)
 
 			if (td->o.use_thread) {
 				dprint(FD_PROCESS, "will pthread_create\n");
-				if (pthread_create(&td->thread, NULL, thread_main, td)) {
-					perror("thread_create");
+				if (pthread_create(&td->thread, NULL,
+						   thread_main, td)) {
+					perror("pthread_create");
 					nr_started--;
 					break;
 				}
+				if (pthread_detach(td->thread) < 0)
+					perror("pthread_detach");
 			} else {
+				pid_t pid;
 				dprint(FD_PROCESS, "will fork\n");
-				if (!fork()) {
+				pid = fork();
+				if (!pid) {
 					int ret = fork_main(shm_id, i);
 
-					exit(ret);
-				}
+					_exit(ret);
+				} else if (i == fio_debug_jobno)
+					*fio_debug_jobp = pid;
 			}
-			fio_sem_down(startup_sem);
+			fio_mutex_down(startup_mutex);
 		}
 
 		/*
@@ -1258,13 +1376,16 @@ static void run_threads(void)
 			if (td->runstate != TD_INITIALIZED)
 				continue;
 
-			td_set_runstate(td, TD_RUNNING);
+			if (in_ramp_time(td))
+				td_set_runstate(td, TD_RAMP);
+			else
+				td_set_runstate(td, TD_RUNNING);
 			nr_running++;
 			nr_started--;
 			m_rate += td->o.ratemin;
 			t_rate += td->o.rate;
 			todo--;
-			fio_sem_up(td->mutex);
+			fio_mutex_up(td->mutex);
 		}
 
 		reap_threads(&nr_running, &t_rate, &m_rate);
@@ -1286,6 +1407,8 @@ int main(int argc, char *argv[])
 {
 	long ps;
 
+	sinit();
+
 	/*
 	 * We need locale for number printing, if it isn't set then just
 	 * go with the US format.
@@ -1313,22 +1436,23 @@ int main(int argc, char *argv[])
 		setup_log(&agg_io_log[DDIR_WRITE]);
 	}
 
-	startup_sem = fio_sem_init(0);
+	startup_mutex = fio_mutex_init(0);
 
 	set_genesis_time();
 
-	disk_util_timer_arm();
+	status_timer_arm();
 
 	run_threads();
 
 	if (!fio_abort) {
 		show_run_stats();
 		if (write_bw_log) {
-			__finish_log(agg_io_log[DDIR_READ],"agg-read_bw.log");
-			__finish_log(agg_io_log[DDIR_WRITE],"agg-write_bw.log");
+			__finish_log(agg_io_log[DDIR_READ], "agg-read_bw.log");
+			__finish_log(agg_io_log[DDIR_WRITE],
+					"agg-write_bw.log");
 		}
 	}
 
-	fio_sem_remove(startup_sem);
+	fio_mutex_remove(startup_mutex);
 	return exit_value;
 }
