@@ -16,17 +16,23 @@
 #include <assert.h>
 
 #include "compiler/compiler.h"
-#include "list.h"
+#include "flist.h"
 #include "fifo.h"
 #include "rbtree.h"
 #include "arch/arch.h"
 #include "os/os.h"
 #include "mutex.h"
+#include "log.h"
+#include "debug.h"
 
 #ifdef FIO_HAVE_GUASI
 #include <guasi.h>
 #endif
 
+#ifdef FIO_HAVE_SOLARISAIO
+#include <sys/asynch.h>
+#endif
+
 enum fio_ddir {
 	DDIR_READ = 0,
 	DDIR_WRITE,
@@ -44,6 +50,12 @@ enum td_ddir {
 	TD_DDIR_RANDRW		= TD_DDIR_RW | TD_DDIR_RAND,
 };
 
+enum file_lock_mode {
+	FILE_LOCK_NONE,
+	FILE_LOCK_EXCLUSIVE,
+	FILE_LOCK_READWRITE,
+};
+
 /*
  * Use for maintaining statistics
  */
@@ -80,7 +92,7 @@ struct io_log {
 struct io_piece {
 	union {
 		struct rb_node rb_node;
-		struct list_head list;
+		struct flist_head list;
 	};
 	union {
 		int fileno;
@@ -118,6 +130,9 @@ struct io_u {
 #endif
 #ifdef FIO_HAVE_GUASI
 		guasi_req_t greq;
+#endif
+#ifdef FIO_HAVE_SOLARISAIO
+		aio_result_t resultp;
 #endif
 	};
 	struct timeval start_time;
@@ -149,13 +164,14 @@ struct io_u {
 	union {
 		unsigned int index;
 		unsigned int seen;
+		void *engine_data;
 	};
 
 	unsigned int flags;
 
 	struct fio_file *file;
 
-	struct list_head list;
+	struct flist_head list;
 
 	/*
 	 * Callback for io completion
@@ -179,6 +195,8 @@ enum {
 	VERIFY_MD5,			/* md5 sum data blocks */
 	VERIFY_CRC64,			/* crc64 sum data blocks */
 	VERIFY_CRC32,			/* crc32 sum data blocks */
+	VERIFY_CRC32C,			/* crc32c sum data blocks */
+	VERIFY_CRC32C_INTEL,		/* crc32c sum data blocks with hw */
 	VERIFY_CRC16,			/* crc16 sum data blocks */
 	VERIFY_CRC7,			/* crc7 sum data blocks */
 	VERIFY_SHA256,			/* sha256 sum data blocks */
@@ -269,11 +287,10 @@ enum fio_ioengine_flags {
 enum fio_file_flags {
 	FIO_FILE_OPEN		= 1 << 0,	/* file is open */
 	FIO_FILE_CLOSING	= 1 << 1,	/* file being closed */
-	FIO_FILE_EXISTS		= 1 << 2,	/* file there */
-	FIO_FILE_EXTEND		= 1 << 3,	/* needs extend */
-	FIO_FILE_NOSORT		= 1 << 4,	/* don't sort verify blocks */
-	FIO_FILE_DONE		= 1 << 5,	/* io completed to this file */
-	FIO_SIZE_KNOWN		= 1 << 6,	/* size has been set */
+	FIO_FILE_EXTEND		= 1 << 2,	/* needs extend */
+	FIO_FILE_DONE		= 1 << 3,	/* io completed to this file */
+	FIO_SIZE_KNOWN		= 1 << 4,	/* size has been set */
+	FIO_FILE_HASHED		= 1 << 5,	/* file is on hash */
 };
 
 /*
@@ -281,6 +298,7 @@ enum fio_file_flags {
  * this structure holds state information for a single file.
  */
 struct fio_file {
+	struct flist_head hash_list;
 	enum fio_filetype filetype;
 
 	/*
@@ -306,12 +324,19 @@ struct fio_file {
 	unsigned long long io_size;
 
 	unsigned long long last_pos;
-	unsigned long long last_completed_pos;
+
+	/*
+	 * if io is protected by a semaphore, this is set
+	 */
+	struct fio_mutex *lock;
+	void *lock_owner;
+	unsigned int lock_batch;
+	enum fio_ddir lock_ddir;
 
 	/*
 	 * block map for random io
 	 */
-	unsigned long *file_map;
+	unsigned int *file_map;
 	unsigned int num_maps;
 	unsigned int last_free_lookup;
 
@@ -363,10 +388,14 @@ struct thread_stat {
 	 * IO depth and latency stats
 	 */
 	unsigned int io_u_map[FIO_IO_U_MAP_NR];
+	unsigned int io_u_submit[FIO_IO_U_MAP_NR];
+	unsigned int io_u_complete[FIO_IO_U_MAP_NR];
 	unsigned int io_u_lat_u[FIO_IO_U_LAT_U_NR];
 	unsigned int io_u_lat_m[FIO_IO_U_LAT_M_NR];
 	unsigned long total_io_u[2];
 	unsigned long short_io_u[2];
+	unsigned long total_submit;
+	unsigned long total_complete;
 
 	unsigned long long io_bytes[2];
 	unsigned long runtime[2];
@@ -391,6 +420,7 @@ struct thread_options {
 	unsigned int iodepth;
 	unsigned int iodepth_low;
 	unsigned int iodepth_batch;
+	unsigned int iodepth_batch_complete;
 
 	unsigned long long size;
 	unsigned int fill_device;
@@ -406,6 +436,8 @@ struct thread_options {
 
 	unsigned int nr_files;
 	unsigned int open_files;
+	enum file_lock_mode file_lock_mode;
+	unsigned int lockfile_batch;
 
 	unsigned int odirect;
 	unsigned int invalidate_cache;
@@ -429,6 +461,7 @@ struct thread_options {
 	unsigned int write_lat_log;
 	unsigned int write_bw_log;
 	unsigned int norandommap;
+	unsigned int softrandommap;
 	unsigned int bs_unaligned;
 	unsigned int fsync_on_close;
 
@@ -440,6 +473,7 @@ struct thread_options {
 	unsigned int fsync_blocks;
 	unsigned int start_delay;
 	unsigned long long timeout;
+	unsigned long long ramp_time;
 	unsigned int overwrite;
 	unsigned int bw_avg_time;
 	unsigned int loops;
@@ -460,6 +494,7 @@ struct thread_options {
 	unsigned int group_reporting;
 	unsigned int fadvise_hint;
 	unsigned int zero_buffers;
+	unsigned int refill_buffers;
 	unsigned int time_based;
 
 	char *read_iolog_file;
@@ -498,7 +533,7 @@ struct thread_data {
 	int thread_number;
 	int groupid;
 	struct thread_stat ts;
-	struct fio_file *files;
+	struct fio_file **files;
 	unsigned int files_index;
 	unsigned int nr_open_files;
 	unsigned int nr_done_files;
@@ -541,10 +576,10 @@ struct thread_data {
 	 * Current IO depth and list of free and busy io_u's.
 	 */
 	unsigned int cur_depth;
-	struct list_head io_u_freelist;
-	struct list_head io_u_busylist;
-	struct list_head io_u_requeues;
 	unsigned int io_u_queued;
+	struct flist_head io_u_freelist;
+	struct flist_head io_u_busylist;
+	struct flist_head io_u_requeues;
 
 	/*
 	 * Rate state
@@ -563,7 +598,7 @@ struct thread_data {
 	unsigned long long io_skip_bytes;
 	unsigned long long this_io_bytes[2];
 	unsigned long long zone_bytes;
-	struct fio_sem *mutex;
+	struct fio_mutex *mutex;
 
 	/*
 	 * State for random io, a bitmap of blocks done vs not done
@@ -575,13 +610,13 @@ struct thread_data {
 	struct timeval rw_end[2];
 	struct timeval last_issue;
 	unsigned int rw_end_set[2];
+	unsigned int ramp_time_over;
 
 	/*
 	 * read/write mixed workload state
 	 */
 	os_random_state_t rwmix_state;
-	unsigned long long rwmix_bytes;
-	struct timeval rwmix_switch;
+	unsigned long rwmix_issues;
 	enum fio_ddir rwmix_ddir;
 	unsigned int ddir_nr;
 
@@ -590,12 +625,12 @@ struct thread_data {
 	 * if we are overwriting. Otherwise just use a fifo.
 	 */
 	struct rb_root io_hist_tree;
-	struct list_head io_hist_list;
+	struct flist_head io_hist_list;
 
 	/*
 	 * For IO replaying
 	 */
-	struct list_head io_log_list;
+	struct flist_head io_log_list;
 
 	/*
 	 * timeout handling
@@ -661,13 +696,12 @@ extern int nr_process, nr_thread;
 extern int shm_id;
 extern int groupid;
 extern int terse_output;
-extern FILE *f_out;
-extern FILE *f_err;
 extern int temp_stall_ts;
 extern unsigned long long mlock_size;
 extern unsigned long page_mask, page_size;
 extern int read_only;
 extern int eta_print;
+extern unsigned long done_secs;
 extern char *job_section;
 
 extern struct thread_data *threads;
@@ -676,16 +710,17 @@ extern struct thread_data *threads;
 #define td_write(td)		((td)->o.td_ddir & TD_DDIR_WRITE)
 #define td_rw(td)		(((td)->o.td_ddir & TD_DDIR_RW) == TD_DDIR_RW)
 #define td_random(td)		((td)->o.td_ddir & TD_DDIR_RAND)
+#define file_randommap(td, f)	(!(td)->o.norandommap && (f)->file_map)
 
 static inline void fio_ro_check(struct thread_data *td, struct io_u *io_u)
 {
 	assert(!(io_u->ddir == DDIR_WRITE && !td_write(td)));
 }
 
-#define BLOCKS_PER_MAP		(8 * sizeof(long))
-#define TO_MAP_BLOCK(td, f, b)	(b)
-#define RAND_MAP_IDX(td, f, b)	(TO_MAP_BLOCK(td, f, b) / BLOCKS_PER_MAP)
-#define RAND_MAP_BIT(td, f, b)	(TO_MAP_BLOCK(td, f, b) & (BLOCKS_PER_MAP - 1))
+#define BLOCKS_PER_MAP		(8 * sizeof(int))
+#define TO_MAP_BLOCK(f, b)	(b)
+#define RAND_MAP_IDX(f, b)	(TO_MAP_BLOCK(f, b) / BLOCKS_PER_MAP)
+#define RAND_MAP_BIT(f, b)	(TO_MAP_BLOCK(f, b) & (BLOCKS_PER_MAP - 1))
 
 #define MAX_JOBS	(1024)
 
@@ -717,7 +752,11 @@ struct disk_util_stat {
  * Per-device disk util management
  */
 struct disk_util {
-	struct list_head list;
+	struct flist_head list;
+	/* If this disk is a slave, hook it into the master's
+	 * list using this head.
+	 */
+	struct flist_head slavelist;
 
 	char *name;
 	char *sysfs_root;
@@ -727,6 +766,15 @@ struct disk_util {
 	struct disk_util_stat dus;
 	struct disk_util_stat last_dus;
 
+	/* For software raids, this entry maintains pointers to the
+	 * entries for the slave devices. The disk_util entries for
+	 * the slaves devices should primarily be maintained through
+	 * the disk_list list, i.e. for memory allocation and
+	 * de-allocation, etc. Whereas this list should be used only
+	 * for aggregating a software RAID's disk util figures.
+	 */
+	struct flist_head slaves;
+
 	unsigned long msec;
 	struct timeval time;
 };
@@ -740,6 +788,7 @@ enum file_log_act {
 	FIO_LOG_ADD_FILE,
 	FIO_LOG_OPEN_FILE,
 	FIO_LOG_CLOSE_FILE,
+	FIO_LOG_UNLINK_FILE,
 };
 
 extern int __must_check read_iolog_get(struct thread_data *, struct io_u *);
@@ -747,6 +796,7 @@ extern void log_io_u(struct thread_data *, struct io_u *);
 extern void log_file(struct thread_data *, struct fio_file *, enum file_log_act);
 extern int __must_check init_iolog(struct thread_data *td);
 extern void log_io_piece(struct thread_data *, struct io_u *);
+extern void queue_io_piece(struct thread_data *, struct io_piece *);
 extern void prune_io_piece_log(struct thread_data *);
 extern void write_iolog_close(struct thread_data *);
 
@@ -760,7 +810,6 @@ extern void show_run_stats(void);
 extern void init_disk_util(struct thread_data *);
 extern void update_rusage_stat(struct thread_data *);
 extern void update_io_ticks(void);
-extern void disk_util_timer_arm(void);
 extern void setup_log(struct io_log **);
 extern void finish_log(struct thread_data *, struct io_log *, const char *);
 extern void __finish_log(struct io_log *, const char *);
@@ -783,12 +832,14 @@ extern void rate_throttle(struct thread_data *, unsigned long, unsigned int);
 extern void fill_start_time(struct timeval *);
 extern void fio_gettime(struct timeval *, void *);
 extern void set_genesis_time(void);
+extern int ramp_time_over(struct thread_data *);
+extern int in_ramp_time(struct thread_data *);
 
 /*
  * Init/option functions
  */
 extern int __must_check parse_options(int, char **);
-extern int fio_option_parse(struct thread_data *, const char *);
+extern int fio_options_parse(struct thread_data *, char **, int);
 extern int fio_cmd_option_parse(struct thread_data *, const char *, char *);
 extern void fio_fill_default_options(struct thread_data *);
 extern int fio_show_option_help(const char *);
@@ -802,14 +853,18 @@ extern void options_mem_free(struct thread_data *);
  * File setup/shutdown
  */
 extern void close_files(struct thread_data *);
+extern void close_and_free_files(struct thread_data *);
 extern int __must_check setup_files(struct thread_data *);
 extern int __must_check open_files(struct thread_data *);
 extern int __must_check file_invalidate_cache(struct thread_data *, struct fio_file *);
 extern int __must_check generic_open_file(struct thread_data *, struct fio_file *);
-extern void generic_close_file(struct thread_data *, struct fio_file *);
+extern int __must_check generic_close_file(struct thread_data *, struct fio_file *);
 extern int add_file(struct thread_data *, const char *);
 extern void get_file(struct fio_file *);
-extern void put_file(struct thread_data *, struct fio_file *);
+extern int __must_check put_file(struct thread_data *, struct fio_file *);
+extern void lock_file(struct thread_data *, struct fio_file *, enum fio_ddir);
+extern void unlock_file(struct thread_data *, struct fio_file *);
+extern void unlock_file_all(struct thread_data *, struct fio_file *);
 extern int add_dir_files(struct thread_data *, const char *);
 extern int init_random_map(struct thread_data *);
 extern void dup_files(struct thread_data *, struct thread_data *);
@@ -827,12 +882,10 @@ extern void print_status_init(int);
  */
 #ifdef FIO_HAVE_DISK_UTIL
 extern void show_disk_util(void);
-extern void disk_util_timer_arm(void);
 extern void init_disk_util(struct thread_data *);
 extern void update_io_ticks(void);
 #else
 #define show_disk_util()
-#define disk_util_timer_arm()
 #define init_disk_util(td)
 #define update_io_ticks()
 #endif
@@ -847,6 +900,7 @@ enum {
 	TD_NOT_CREATED = 0,
 	TD_CREATED,
 	TD_INITIALIZED,
+	TD_RAMP,
 	TD_RUNNING,
 	TD_VERIFYING,
 	TD_FSYNCING,
@@ -854,6 +908,8 @@ enum {
 	TD_REAPED,
 };
 
+extern void td_set_runstate(struct thread_data *, int);
+
 /*
  * Verify helpers
  */
@@ -872,7 +928,7 @@ extern void free_io_mem(struct thread_data *);
 /*
  * io unit handling
  */
-#define queue_full(td)	list_empty(&(td)->io_u_freelist)
+#define queue_full(td)	flist_empty(&(td)->io_u_freelist)
 extern struct io_u *__get_io_u(struct thread_data *);
 extern struct io_u *get_io_u(struct thread_data *);
 extern void put_io_u(struct thread_data *, struct io_u *);
@@ -883,7 +939,15 @@ extern void io_u_queued(struct thread_data *, struct io_u *);
 extern void io_u_log_error(struct thread_data *, struct io_u *);
 extern void io_u_init_timeout(void);
 extern void io_u_set_timeout(struct thread_data *);
-extern void io_u_mark_depth(struct thread_data *, struct io_u *);
+extern void io_u_mark_depth(struct thread_data *, unsigned int);
+extern void io_u_fill_buffer(struct thread_data *td, struct io_u *, unsigned int);
+void io_u_mark_complete(struct thread_data *, unsigned int);
+void io_u_mark_submit(struct thread_data *, unsigned int);
+
+/*
+ * Reset stats after ramp time completes
+ */
+extern void reset_all_stats(struct thread_data *);
 
 /*
  * io engine entry points
@@ -895,7 +959,7 @@ extern int __must_check td_io_sync(struct thread_data *, struct fio_file *);
 extern int __must_check td_io_getevents(struct thread_data *, unsigned int, unsigned int, struct timespec *);
 extern int __must_check td_io_commit(struct thread_data *);
 extern int __must_check td_io_open_file(struct thread_data *, struct fio_file *);
-extern void td_io_close_file(struct thread_data *, struct fio_file *);
+extern int td_io_close_file(struct thread_data *, struct fio_file *);
 
 /*
  * blktrace support
@@ -905,22 +969,8 @@ extern int is_blktrace(const char *);
 extern int load_blktrace(struct thread_data *, const char *);
 #endif
 
-/*
- * If logging output to a file, stderr should go to both stderr and f_err
- */
-#define log_err(args...)	do {		\
-	fprintf(f_err, ##args);			\
-	if (f_err != stderr)			\
-		fprintf(stderr, ##args);	\
-	} while (0)
-
-#define log_info(args...)	fprintf(f_out, ##args)
-
-FILE *get_f_out(void);
-FILE *get_f_err(void);
-
 struct ioengine_ops {
-	struct list_head list;
+	struct flist_head list;
 	char name[16];
 	int version;
 	int flags;
@@ -934,12 +984,12 @@ struct ioengine_ops {
 	int (*cancel)(struct thread_data *, struct io_u *);
 	void (*cleanup)(struct thread_data *);
 	int (*open_file)(struct thread_data *, struct fio_file *);
-	void (*close_file)(struct thread_data *, struct fio_file *);
+	int (*close_file)(struct thread_data *, struct fio_file *);
 	void *data;
 	void *dlhandle;
 };
 
-#define FIO_IOOPS_VERSION	8
+#define FIO_IOOPS_VERSION	9
 
 extern struct ioengine_ops *load_ioengine(struct thread_data *, const char *);
 extern void register_ioengine(struct ioengine_ops *);
@@ -956,7 +1006,10 @@ extern void close_ioengine(struct thread_data *);
 #define for_each_td(td, i)	\
 	for ((i) = 0, (td) = &threads[0]; (i) < (int) thread_number; (i)++, (td)++)
 #define for_each_file(td, f, i)	\
-	for ((i) = 0, (f) = &(td)->files[0]; (i) < (td)->o.nr_files; (i)++, (f)++)
+	if ((td)->files_index)						\
+		for ((i) = 0, (f) = (td)->files[0];			\
+	    	 (i) < (td)->o.nr_files && ((f) = (td)->files[i]) != NULL; \
+		 (i)++)
 
 #define fio_assert(td, cond)	do {	\
 	if (!(cond)) {			\
@@ -968,47 +1021,26 @@ extern void close_ioengine(struct thread_data *);
 	}	\
 } while (0)
 
+static inline void fio_file_reset(struct fio_file *f)
+{
+	f->last_free_lookup = 0;
+	f->last_pos = f->file_offset;
+}
+
 static inline void clear_error(struct thread_data *td)
 {
 	td->error = 0;
 	td->verror[0] = '\0';
 }
 
-enum {
-	FD_PROCESS	= 0,
-	FD_FILE,
-	FD_IO,
-	FD_MEM,
-	FD_BLKTRACE,
-	FD_VERIFY,
-	FD_RANDOM,
-	FD_DEBUG_MAX,
-};
-
 #ifdef FIO_INC_DEBUG
-struct debug_level {
-	const char *name;
-	unsigned long shift;
-};
-extern struct debug_level debug_levels[];
-
-extern unsigned long fio_debug;
-
-#define dprint(type, str, args...)				\
-	do {							\
-		assert(type < FD_DEBUG_MAX);			\
-		if ((((1 << type)) & fio_debug) == 0)		\
-			break;					\
-		log_info("%-8s ", debug_levels[(type)].name);	\
-		log_info(str, ##args);				\
-	} while (0)
-
 static inline void dprint_io_u(struct io_u *io_u, const char *p)
 {
 	struct fio_file *f = io_u->file;
 
 	dprint(FD_IO, "%s: io_u %p: off=%llu/len=%lu/ddir=%d", p, io_u,
-					io_u->offset, io_u->buflen, io_u->ddir);
+					(unsigned long long) io_u->offset,
+					io_u->buflen, io_u->ddir);
 	if (fio_debug & (1 << FD_IO)) {
 		if (f)
 			log_info("/%s", f->file_name);
@@ -1017,7 +1049,6 @@ static inline void dprint_io_u(struct io_u *io_u, const char *p)
 	}
 }
 #else
-#define dprint(type, str, args...)
 #define dprint_io_u(io_u, p)
 #endif
 
