@@ -16,8 +16,10 @@
 
 #include "fio.h"
 #include "parse.h"
+#include "smalloc.h"
+#include "filehash.h"
 
-static char fio_version_string[] = "fio 1.19";
+static char fio_version_string[] = "fio 1.22";
 
 #define FIO_RANDSEED		(0xb1899bedUL)
 
@@ -39,18 +41,20 @@ char *job_section = NULL;
 int write_bw_log = 0;
 int read_only = 0;
 
-static int def_timeout = 0;
-static int write_lat_log = 0;
+static int def_timeout;
+static int write_lat_log;
 
 static int prev_group_jobs;
 
 unsigned long fio_debug = 0;
+unsigned int fio_debug_jobno = -1;
+unsigned int *fio_debug_jobp = NULL;
 
 /*
  * Command line options. These will contain the above, plus a few
  * extra that only pertain to fio itself and not jobs.
  */
-static struct option long_options[FIO_NR_OPTIONS] = {
+static struct option l_opts[FIO_NR_OPTIONS] = {
 	{
 		.name		= "output",
 		.has_arg	= required_argument,
@@ -116,6 +120,11 @@ static struct option long_options[FIO_NR_OPTIONS] = {
 		.has_arg	= required_argument,
 		.val		= 'x',
 	},
+	{
+		.name		= "alloc-size",
+		.has_arg	= required_argument,
+		.val		= 'a',
+	},
 	{
 		.name		= NULL,
 	},
@@ -206,12 +215,10 @@ static int fixup_options(struct thread_data *td)
 	struct thread_options *o = &td->o;
 
 	if (read_only && td_write(td)) {
-		log_err("fio: job <%s> has write bit set, but fio is in read-only mode\n", td->o.name);
+		log_err("fio: job <%s> has write bit set, but fio is in"
+			" read-only mode\n", td->o.name);
 		return 1;
 	}
-	
-	if (o->rwmix[DDIR_READ] + o->rwmix[DDIR_WRITE] > 100)
-		o->rwmix[DDIR_WRITE] = 100 - o->rwmix[DDIR_READ];
 
 	if (o->write_iolog_file && o->read_iolog_file) {
 		log_err("fio: read iolog overrides write_iolog\n");
@@ -232,11 +239,11 @@ static int fixup_options(struct thread_data *td)
 		o->overwrite = 1;
 
 	if (!o->min_bs[DDIR_READ])
-		o->min_bs[DDIR_READ]= o->bs[DDIR_READ];
+		o->min_bs[DDIR_READ] = o->bs[DDIR_READ];
 	if (!o->max_bs[DDIR_READ])
 		o->max_bs[DDIR_READ] = o->bs[DDIR_READ];
 	if (!o->min_bs[DDIR_WRITE])
-		o->min_bs[DDIR_WRITE]= o->bs[DDIR_WRITE];
+		o->min_bs[DDIR_WRITE] = o->bs[DDIR_WRITE];
 	if (!o->max_bs[DDIR_WRITE])
 		o->max_bs[DDIR_WRITE] = o->bs[DDIR_WRITE];
 
@@ -300,12 +307,15 @@ static int fixup_options(struct thread_data *td)
 	}
 
 	if (o->fill_device && !o->size)
-		o->size = ULONG_LONG_MAX;
-	
+		o->size = -1ULL;
+
 	if (td_rw(td) && td->o.verify != VERIFY_NONE)
 		log_info("fio: mixed read/write workload with verify. May not "
 		 "work as expected, unless you pre-populated the file\n");
 
+	if (td->o.verify != VERIFY_NONE)
+		td->o.refill_buffers = 1;
+
 	return 0;
 }
 
@@ -443,14 +453,15 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 		td->io_ops->flags |= FIO_RAWIO;
 
 	file_alloced = 0;
-	if (!td->o.filename && !td->files_index) {
+	if (!td->o.filename && !td->files_index && !td->o.read_iolog_file) {
 		file_alloced = 1;
 
 		if (td->o.nr_files == 1 && exists_and_not_file(jobname))
 			add_file(td, jobname);
 		else {
 			for (i = 0; i < td->o.nr_files; i++) {
-				sprintf(fname, "%s.%d.%d", jobname, td->thread_number, i);
+				sprintf(fname, "%s.%d.%d", jobname,
+							td->thread_number, i);
 				add_file(td, fname);
 			}
 		}
@@ -466,7 +477,7 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 			f->real_file_size = -1ULL;
 	}
 
-	td->mutex = fio_sem_init(0);
+	td->mutex = fio_mutex_init(0);
 
 	td->ts.clat_stat[0].min_val = td->ts.clat_stat[1].min_val = ULONG_MAX;
 	td->ts.slat_stat[0].min_val = td->ts.slat_stat[1].min_val = ULONG_MAX;
@@ -500,9 +511,12 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 
 	if (!terse_output) {
 		if (!job_add_num) {
-			if (!strcmp(td->io_ops->name, "cpuio"))
-				log_info("%s: ioengine=cpu, cpuload=%u, cpucycle=%u\n", td->o.name, td->o.cpuload, td->o.cpucycle);
-			else {
+			if (!strcmp(td->io_ops->name, "cpuio")) {
+				log_info("%s: ioengine=cpu, cpuload=%u,"
+					 " cpucycle=%u\n", td->o.name,
+							td->o.cpuload,
+							td->o.cpucycle);
+			} else {
 				char *c1, *c2, *c3, *c4;
 
 				c1 = to_kmg(td->o.min_bs[DDIR_READ]);
@@ -510,7 +524,13 @@ static int add_job(struct thread_data *td, const char *jobname, int job_add_num)
 				c3 = to_kmg(td->o.min_bs[DDIR_WRITE]);
 				c4 = to_kmg(td->o.max_bs[DDIR_WRITE]);
 
-				log_info("%s: (g=%d): rw=%s, bs=%s-%s/%s-%s, ioengine=%s, iodepth=%u\n", td->o.name, td->groupid, ddir_str[td->o.td_ddir], c1, c2, c3, c4, td->io_ops->name, td->o.iodepth);
+				log_info("%s: (g=%d): rw=%s, bs=%s-%s/%s-%s,"
+					 " ioengine=%s, iodepth=%u\n",
+						td->o.name, td->groupid,
+						ddir_str[td->o.td_ddir],
+						c1, c2, c3, c4,
+						td->io_ops->name,
+						td->o.iodepth);
 
 				free(c1);
 				free(c2);
@@ -594,6 +614,8 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 	int first_sect = 1;
 	int skip_fgets = 0;
 	int inside_skip = 0;
+	char **opts;
+	int i, alloc_opts, num_opts;
 
 	if (!strcmp(file, "-"))
 		f = stdin;
@@ -613,6 +635,10 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 	name = malloc(280);
 	memset(name, 0, 280);
 
+	alloc_opts = 8;
+	opts = malloc(sizeof(char *) * alloc_opts);
+	num_opts = 0;
+
 	stonewall = stonewall_flag;
 	do {
 		/*
@@ -634,7 +660,8 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 		if (sscanf(p, "[%255s]", name) != 1) {
 			if (inside_skip)
 				continue;
-			log_err("fio: option <%s> outside of [] job section\n", p);
+			log_err("fio: option <%s> outside of [] job section\n",
+									p);
 			break;
 		}
 
@@ -670,6 +697,9 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 			stonewall = 0;
 		}
 
+		num_opts = 0;
+		memset(opts, 0, alloc_opts * sizeof(char *));
+
 		while ((p = fgets(string, 4096, f)) != NULL) {
 			if (is_empty_or_comment(p))
 				continue;
@@ -687,29 +717,42 @@ static int parse_jobs_ini(char *file, int stonewall_flag)
 
 			strip_blank_end(p);
 
-			/*
-			 * Don't break here, continue parsing options so we
-			 * dump all the bad ones. Makes trial/error fixups
-			 * easier on the user.
-			 */
-			ret |= fio_option_parse(td, p);
-			if (!ret && dump_cmdline)
-				log_info("--%s ", p);
+			if (num_opts == alloc_opts) {
+				alloc_opts <<= 1;
+				opts = realloc(opts,
+						alloc_opts * sizeof(char *));
+			}
+
+			opts[num_opts] = strdup(p);
+			num_opts++;
 		}
 
-		if (!ret)
+		ret = fio_options_parse(td, opts, num_opts);
+		if (!ret) {
+			if (dump_cmdline)
+				for (i = 0; i < num_opts; i++)
+					log_info("--%s ", opts[i]);
+
 			ret = add_job(td, name, 0);
-		else {
+		} else {
 			log_err("fio: job %s dropped\n", name);
 			put_job(td);
 		}
+
+		for (i = 0; i < num_opts; i++)
+			free(opts[i]);
+		num_opts = 0;
 	} while (!ret);
 
 	if (dump_cmdline)
 		log_info("\n");
 
+	for (i = 0; i < num_opts; i++)
+		free(opts[i]);
+
 	free(string);
 	free(name);
+	free(opts);
 	if (f != stdin)
 		fclose(f);
 	return ret;
@@ -738,19 +781,28 @@ static void free_shm(void)
 	struct shmid_ds sbuf;
 
 	if (threads) {
-		shmdt((void *) threads);
+		void *tp = threads;
+
 		threads = NULL;
+		file_hash_exit();
+		fio_debug_jobp = NULL;
+		shmdt(tp);
 		shmctl(shm_id, IPC_RMID, &sbuf);
 	}
+
+	scleanup();
 }
 
 /*
  * The thread area is shared between the main process and the job
  * threads/processes. So setup a shared memory segment that will hold
- * all the job info.
+ * all the job info. We use the end of the region for keeping track of
+ * open files across jobs, for file sharing.
  */
 static int setup_thread_area(void)
 {
+	void *hash;
+
 	/*
 	 * 1024 is too much on some machines, scale max_jobs if
 	 * we get a failure that looks like too large a shm segment
@@ -758,6 +810,9 @@ static int setup_thread_area(void)
 	do {
 		size_t size = max_jobs * sizeof(struct thread_data);
 
+		size += file_hash_size;
+		size += sizeof(unsigned int);
+
 		shm_id = shmget(0, size, IPC_CREAT | 0600);
 		if (shm_id != -1)
 			break;
@@ -779,6 +834,10 @@ static int setup_thread_area(void)
 	}
 
 	memset(threads, 0, max_jobs * sizeof(struct thread_data));
+	hash = (void *) threads + max_jobs * sizeof(struct thread_data);
+	fio_debug_jobp = (void *) hash + file_hash_size;
+	*fio_debug_jobp = -1;
+	file_hash_init(hash);
 	atexit(free_shm);
 	return 0;
 }
@@ -795,12 +854,16 @@ static void usage(const char *name)
 	printf("\t--minimal\tMinimal (terse) output\n");
 	printf("\t--version\tPrint version info and exit\n");
 	printf("\t--help\t\tPrint this page\n");
-	printf("\t--cmdhelp=cmd\tPrint command help, \"all\" for all of them\n");
+	printf("\t--cmdhelp=cmd\tPrint command help, \"all\" for all of"
+		" them\n");
 	printf("\t--showcmd\tTurn a job file into command line options\n");
 	printf("\t--eta=when\tWhen ETA estimate should be printed\n");
 	printf("\t          \tMay be \"always\", \"never\" or \"auto\"\n");
-	printf("\t--readonly\tTurn on safety read-only checks, preventing writes\n");
+	printf("\t--readonly\tTurn on safety read-only checks, preventing"
+		" writes\n");
 	printf("\t--section=name\tOnly run specified section in job file\n");
+	printf("\t--alloc-size=kb\tSet smalloc pool to this size in kb"
+		" (def 1024)\n");
 }
 
 #ifdef FIO_INC_DEBUG
@@ -812,7 +875,10 @@ struct debug_level debug_levels[] = {
 	{ .name = "blktrace",	.shift = FD_BLKTRACE },
 	{ .name = "verify",	.shift = FD_VERIFY },
 	{ .name = "random",	.shift = FD_RANDOM },
-	{ },
+	{ .name = "parse",	.shift = FD_PARSE },
+	{ .name = "diskutil",	.shift = FD_DISKUTIL },
+	{ .name = "job",	.shift = FD_JOB },
+	{ .name = NULL, },
 };
 
 static int set_debug(const char *string)
@@ -832,22 +898,38 @@ static int set_debug(const char *string)
 		}
 		log_info("all\n");
 		return 1;
-	} else if (!strcmp(string, "all")) {
-		fio_debug = ~0UL;
-		return 0;
 	}
 
 	while ((opt = strsep(&p, ",")) != NULL) {
 		int found = 0;
 
+		if (!strncmp(opt, "all", 3)) {
+			log_info("fio: set all debug options\n");
+			fio_debug = ~0UL;
+			continue;
+		}
+
 		for (i = 0; debug_levels[i].name; i++) {
 			dl = &debug_levels[i];
-			if (!strncmp(opt, dl->name, strlen(opt))) {
+			found = !strncmp(opt, dl->name, strlen(dl->name));
+			if (!found)
+				continue;
+
+			if (dl->shift == FD_JOB) {
+				opt = strchr(opt, ':');
+				if (!opt) {
+					log_err("fio: missing job number\n");
+					break;
+				}
+				opt++;
+				fio_debug_jobno = atoi(opt);
+				log_info("fio: set debug jobno %d\n",
+							fio_debug_jobno);
+			} else {
 				log_info("fio: set debug option %s\n", opt);
-				found = 1;
 				fio_debug |= (1UL << dl->shift);
-				break;
 			}
+			break;
 		}
 
 		if (!found)
@@ -856,7 +938,7 @@ static int set_debug(const char *string)
 	return 0;
 }
 #else
-static void set_debug(const char *string)
+static int set_debug(const char *string)
 {
 	log_err("fio: debug tracing not included in build\n");
 	return 1;
@@ -868,8 +950,11 @@ static int parse_cmd_line(int argc, char *argv[])
 	struct thread_data *td = NULL;
 	int c, ini_idx = 0, lidx, ret = 0, do_exit = 0, exit_val = 0;
 
-	while ((c = getopt_long_only(argc, argv, "", long_options, &lidx)) != -1) {
+	while ((c = getopt_long_only(argc, argv, "", l_opts, &lidx)) != -1) {
 		switch (c) {
+		case 'a':
+			smalloc_pool_size = atoi(optarg);
+			break;
 		case 't':
 			def_timeout = atoi(optarg);
 			break;
@@ -916,7 +1001,8 @@ static int parse_cmd_line(int argc, char *argv[])
 			break;
 		case 'x':
 			if (!strcmp(optarg, "global")) {
-				log_err("fio: can't use global as only section\n");
+				log_err("fio: can't use global as only "
+					"section\n");
 				do_exit++;
 				exit_val = 1;
 				break;
@@ -926,7 +1012,7 @@ static int parse_cmd_line(int argc, char *argv[])
 			job_section = strdup(optarg);
 			break;
 		case FIO_GETOPT_JOB: {
-			const char *opt = long_options[lidx].name;
+			const char *opt = l_opts[lidx].name;
 			char *val = optarg;
 
 			if (!strncmp(opt, "name", 4) && td) {
@@ -989,7 +1075,7 @@ int parse_options(int argc, char *argv[])
 	f_out = stdout;
 	f_err = stderr;
 
-	fio_options_dup_and_init(long_options);
+	fio_options_dup_and_init(l_opts);
 
 	if (setup_thread_area())
 		return 1;
