@@ -6,6 +6,7 @@
 #include <assert.h>
 
 #include "fio.h"
+#include "hash.h"
 
 /*
  * Change this define to play with the timeout handling
@@ -24,15 +25,14 @@ struct io_completion_data {
  * The ->file_map[] contains a map of blocks we have or have not done io
  * to yet. Used to make sure we cover the entire range in a fair fashion.
  */
-static int random_map_free(struct thread_data *td, struct fio_file *f,
-			   const unsigned long long block)
+static int random_map_free(struct fio_file *f, const unsigned long long block)
 {
-	unsigned int idx = RAND_MAP_IDX(td, f, block);
-	unsigned int bit = RAND_MAP_BIT(td, f, block);
+	unsigned int idx = RAND_MAP_IDX(f, block);
+	unsigned int bit = RAND_MAP_BIT(f, block);
 
 	dprint(FD_RANDOM, "free: b=%llu, idx=%u, bit=%u\n", block, idx, bit);
 
-	return (f->file_map[idx] & (1UL << bit)) == 0;
+	return (f->file_map[idx] & (1 << bit)) == 0;
 }
 
 /*
@@ -43,48 +43,68 @@ static void mark_random_map(struct thread_data *td, struct io_u *io_u)
 	unsigned int min_bs = td->o.rw_min_bs;
 	struct fio_file *f = io_u->file;
 	unsigned long long block;
-	unsigned int blocks;
-	unsigned int nr_blocks;
+	unsigned int blocks, nr_blocks;
 
 	block = (io_u->offset - f->file_offset) / (unsigned long long) min_bs;
-	blocks = 0;
 	nr_blocks = (io_u->buflen + min_bs - 1) / min_bs;
+	blocks = 0;
 
-	while (blocks < nr_blocks) {
+	while (nr_blocks) {
+		unsigned int this_blocks, mask;
 		unsigned int idx, bit;
 
 		/*
 		 * If we have a mixed random workload, we may
 		 * encounter blocks we already did IO to.
 		 */
-		if ((td->o.ddir_nr == 1) && !random_map_free(td, f, block))
+		if ((td->o.ddir_nr == 1) && !random_map_free(f, block)) {
+			if (!blocks)
+				blocks = 1;
 			break;
+		}
 
-		idx = RAND_MAP_IDX(td, f, block);
-		bit = RAND_MAP_BIT(td, f, block);
+		idx = RAND_MAP_IDX(f, block);
+		bit = RAND_MAP_BIT(f, block);
 
 		fio_assert(td, idx < f->num_maps);
 
-		f->file_map[idx] |= (1UL << bit);
-		block++;
-		blocks++;
+		this_blocks = nr_blocks;
+		if (this_blocks + bit > BLOCKS_PER_MAP)
+			this_blocks = BLOCKS_PER_MAP - bit;
+
+		if (this_blocks == BLOCKS_PER_MAP)
+			mask = -1U;
+		else
+			mask = ((1U << this_blocks) - 1) << bit;
+
+		f->file_map[idx] |= mask;
+		nr_blocks -= this_blocks;
+		blocks += this_blocks;
+		block += this_blocks;
 	}
 
 	if ((blocks * min_bs) < io_u->buflen)
 		io_u->buflen = blocks * min_bs;
 }
 
-static inline unsigned long long last_block(struct thread_data *td,
-					    struct fio_file *f,
-					    enum fio_ddir ddir)
+static unsigned long long last_block(struct thread_data *td, struct fio_file *f,
+				     enum fio_ddir ddir)
 {
 	unsigned long long max_blocks;
+	unsigned long long max_size;
 
-	max_blocks = f->io_size / (unsigned long long) td->o.min_bs[ddir];
+	/*
+	 * Hmm, should we make sure that ->io_size <= ->real_file_size?
+	 */
+	max_size = f->io_size;
+	if (max_size > f->real_file_size)
+		max_size = f->real_file_size;
+
+	max_blocks = max_size / (unsigned long long) td->o.min_bs[ddir];
 	if (!max_blocks)
 		return 0;
 
-	return max_blocks - 1;
+	return max_blocks;
 }
 
 /*
@@ -99,8 +119,8 @@ static int get_next_free_block(struct thread_data *td, struct fio_file *f,
 	i = f->last_free_lookup;
 	*b = (i * BLOCKS_PER_MAP);
 	while ((*b) * min_bs < f->real_file_size) {
-		if (f->file_map[i] != -1UL) {
-			*b += fio_ffz(f->file_map[i]);
+		if (f->file_map[i] != (unsigned int) -1) {
+			*b += ffz(f->file_map[i]);
 			if (*b > last_block(td, f, ddir))
 				break;
 			f->last_free_lookup = i;
@@ -124,18 +144,19 @@ static int get_next_rand_offset(struct thread_data *td, struct fio_file *f,
 	do {
 		r = os_random_long(&td->random_state);
 		dprint(FD_RANDOM, "off rand %llu\n", r);
-		*b = (last_block(td, f, ddir) - 1) * (r / ((unsigned long long) RAND_MAX + 1.0));
+		*b = (last_block(td, f, ddir) - 1)
+			* (r / ((unsigned long long) OS_RAND_MAX + 1.0));
 
 		/*
 		 * if we are not maintaining a random map, we are done.
 		 */
-		if (td->o.norandommap)
+		if (!file_randommap(td, f))
 			return 0;
 
 		/*
 		 * calculate map offset and check if it's free
 		 */
-		if (random_map_free(td, f, *b))
+		if (random_map_free(f, *b))
 			return 0;
 
 		dprint(FD_RANDOM, "get_next_rand_offset: offset %llu busy\n",
@@ -149,7 +170,8 @@ static int get_next_rand_offset(struct thread_data *td, struct fio_file *f,
 	 */
 	loops = 10;
 	do {
-		f->last_free_lookup = (f->num_maps - 1) * (r / (RAND_MAX+1.0));
+		f->last_free_lookup = (f->num_maps - 1) *
+					(r / (OS_RAND_MAX + 1.0));
 		if (!get_next_free_block(td, f, ddir, b))
 			return 0;
 
@@ -188,7 +210,14 @@ static int get_next_offset(struct thread_data *td, struct io_u *io_u)
 			b = (f->last_pos - f->file_offset) / td->o.min_bs[ddir];
 	}
 
-	io_u->offset = (b * td->o.min_bs[ddir]) + f->file_offset;
+	io_u->offset = b * td->o.min_bs[ddir];
+	if (io_u->offset >= f->io_size) {
+		dprint(FD_IO, "get_next_offset: offset %llu >= io_size %llu\n",
+					io_u->offset, f->io_size);
+		return 1;
+	}
+
+	io_u->offset += f->file_offset;
 	if (io_u->offset >= f->real_file_size) {
 		dprint(FD_IO, "get_next_offset: offset %llu >= size %llu\n",
 					io_u->offset, f->real_file_size);
@@ -198,19 +227,31 @@ static int get_next_offset(struct thread_data *td, struct io_u *io_u)
 	return 0;
 }
 
+static inline int is_power_of_2(unsigned int val)
+{
+	return (val != 0 && ((val & (val - 1)) == 0));
+}
+
 static unsigned int get_next_buflen(struct thread_data *td, struct io_u *io_u)
 {
 	const int ddir = io_u->ddir;
-	unsigned int buflen;
+	unsigned int uninitialized_var(buflen);
+	unsigned int minbs, maxbs;
 	long r;
 
-	if (td->o.min_bs[ddir] == td->o.max_bs[ddir])
-		buflen = td->o.min_bs[ddir];
+	minbs = td->o.min_bs[ddir];
+	maxbs = td->o.max_bs[ddir];
+
+	if (minbs == maxbs)
+		buflen = minbs;
 	else {
 		r = os_random_long(&td->bsrange_state);
-		if (!td->o.bssplit_nr)
-			buflen = (unsigned int) (1 + (double) (td->o.max_bs[ddir] - 1) * r / (RAND_MAX + 1.0));
-		else {
+		if (!td->o.bssplit_nr) {
+			buflen = 1 + (unsigned int) ((double) maxbs *
+					(r / (OS_RAND_MAX + 1.0)));
+			if (buflen < minbs)
+				buflen = minbs;
+		} else {
 			long perc = 0;
 			unsigned int i;
 
@@ -219,18 +260,18 @@ static unsigned int get_next_buflen(struct thread_data *td, struct io_u *io_u)
 
 				buflen = bsp->bs;
 				perc += bsp->perc;
-				if (r <= ((LONG_MAX / 100L) * perc))
+				if (r <= ((OS_RAND_MAX / 100L) * perc))
 					break;
 			}
 		}
-		if (!td->o.bs_unaligned)
-			buflen = (buflen + td->o.min_bs[ddir] - 1) & ~(td->o.min_bs[ddir] - 1);
+		if (!td->o.bs_unaligned && is_power_of_2(minbs))
+			buflen = (buflen + minbs - 1) & ~(minbs - 1);
 	}
 
 	if (io_u->offset + buflen > io_u->file->real_file_size) {
 		dprint(FD_IO, "lower buflen %u -> %u (ddir=%d)\n", buflen,
-						td->o.min_bs[ddir], ddir);
-		buflen = td->o.min_bs[ddir];
+						minbs, ddir);
+		buflen = minbs;
 	}
 
 	return buflen;
@@ -238,7 +279,6 @@ static unsigned int get_next_buflen(struct thread_data *td, struct io_u *io_u)
 
 static void set_rwmix_bytes(struct thread_data *td)
 {
-	unsigned long long rbytes;
 	unsigned int diff;
 
 	/*
@@ -246,10 +286,8 @@ static void set_rwmix_bytes(struct thread_data *td)
 	 * buffered writes may issue a lot quicker than they complete,
 	 * whereas reads do not.
 	 */
-	rbytes = td->io_bytes[td->rwmix_ddir] - td->rwmix_bytes;
 	diff = td->o.rwmix[td->rwmix_ddir ^ 1];
-
-	td->rwmix_bytes = td->io_bytes[td->rwmix_ddir] + (rbytes * ((100 - diff)) / diff);
+	td->rwmix_issues = (td->io_issues[td->rwmix_ddir] * diff) / 100;
 }
 
 static inline enum fio_ddir get_rand_ddir(struct thread_data *td)
@@ -258,8 +296,8 @@ static inline enum fio_ddir get_rand_ddir(struct thread_data *td)
 	long r;
 
 	r = os_random_long(&td->rwmix_state);
-	v = 1 + (int) (100.0 * (r / (RAND_MAX + 1.0)));
-	if (v < td->o.rwmix[DDIR_READ])
+	v = 1 + (int) (100.0 * (r / (OS_RAND_MAX + 1.0)));
+	if (v <= td->o.rwmix[DDIR_READ])
 		return DDIR_READ;
 
 	return DDIR_WRITE;
@@ -273,27 +311,12 @@ static inline enum fio_ddir get_rand_ddir(struct thread_data *td)
 static enum fio_ddir get_rw_ddir(struct thread_data *td)
 {
 	if (td_rw(td)) {
-		struct timeval now;
-		unsigned long elapsed;
-		unsigned int cycle;
-
-		fio_gettime(&now, NULL);
-	 	elapsed = mtime_since_now(&td->rwmix_switch);
-
-		/*
-		 * if this is the first cycle, make it shorter
-		 */
-		cycle = td->o.rwmixcycle;
-		if (!td->rwmix_bytes)
-			cycle /= 10;
-
 		/*
 		 * Check if it's time to seed a new data direction.
 		 */
-		if (elapsed >= cycle ||
-		    td->io_bytes[td->rwmix_ddir] >= td->rwmix_bytes) {
+		if (td->io_issues[td->rwmix_ddir] >= td->rwmix_issues) {
 			unsigned long long max_bytes;
-			enum fio_ddir ddir;			
+			enum fio_ddir ddir;
 
 			/*
 			 * Put a top limit on how many bytes we do for
@@ -302,11 +325,13 @@ static enum fio_ddir get_rw_ddir(struct thread_data *td)
 			 */
 			ddir = get_rand_ddir(td);
 			max_bytes = td->this_io_bytes[ddir];
-			if (max_bytes >= (td->o.size * td->o.rwmix[ddir] / 100)) {
+			if (max_bytes >=
+			    (td->o.size * td->o.rwmix[ddir] / 100)) {
 				if (!td->rw_end_set[ddir]) {
 					td->rw_end_set[ddir] = 1;
-					memcpy(&td->rw_end[ddir], &now, sizeof(now));
+					fio_gettime(&td->rw_end[ddir], NULL);
 				}
+
 				ddir ^= 1;
 			}
 
@@ -314,7 +339,6 @@ static enum fio_ddir get_rw_ddir(struct thread_data *td)
 				set_rwmix_bytes(td);
 
 			td->rwmix_ddir = ddir;
-			memcpy(&td->rwmix_switch, &now, sizeof(now));
 		}
 		return td->rwmix_ddir;
 	} else if (td_read(td))
@@ -323,17 +347,25 @@ static enum fio_ddir get_rw_ddir(struct thread_data *td)
 		return DDIR_WRITE;
 }
 
+static void put_file_log(struct thread_data *td, struct fio_file *f)
+{
+	int ret = put_file(td, f);
+
+	if (ret)
+		td_verror(td, ret, "file close");
+}
+
 void put_io_u(struct thread_data *td, struct io_u *io_u)
 {
 	assert((io_u->flags & IO_U_F_FREE) == 0);
 	io_u->flags |= IO_U_F_FREE;
 
 	if (io_u->file)
-		put_file(td, io_u->file);
+		put_file_log(td, io_u->file);
 
 	io_u->file = NULL;
-	list_del(&io_u->list);
-	list_add(&io_u->list, &td->io_u_freelist);
+	flist_del(&io_u->list);
+	flist_add(&io_u->list, &td->io_u_freelist);
 	td->cur_depth--;
 }
 
@@ -341,14 +373,16 @@ void requeue_io_u(struct thread_data *td, struct io_u **io_u)
 {
 	struct io_u *__io_u = *io_u;
 
+	dprint(FD_IO, "requeue %p\n", __io_u);
+
 	__io_u->flags |= IO_U_F_FREE;
 	if ((__io_u->flags & IO_U_F_FLIGHT) && (__io_u->ddir != DDIR_SYNC))
 		td->io_issues[__io_u->ddir]--;
-		
+
 	__io_u->flags &= ~IO_U_F_FLIGHT;
 
-	list_del(&__io_u->list);
-	list_add_tail(&__io_u->list, &td->io_u_requeues);
+	flist_del(&__io_u->list);
+	flist_add_tail(&__io_u->list, &td->io_u_requeues);
 	td->cur_depth--;
 	*io_u = NULL;
 }
@@ -404,7 +438,7 @@ static int fill_io_u(struct thread_data *td, struct io_u *io_u)
 	/*
 	 * mark entry before potentially trimming io_u
 	 */
-	if (td_random(td) && !td->o.norandommap)
+	if (td_random(td) && file_randommap(td, io_u->file))
 		mark_random_map(td, io_u);
 
 	/*
@@ -417,12 +451,50 @@ out:
 	return 0;
 }
 
-void io_u_mark_depth(struct thread_data *td, struct io_u *io_u)
+static void __io_u_mark_map(unsigned int *map, unsigned int nr)
 {
 	int index = 0;
 
-	if (io_u->ddir == DDIR_SYNC)
-		return;
+	switch (nr) {
+	default:
+		index = 6;
+		break;
+	case 33 ... 64:
+		index = 5;
+		break;
+	case 17 ... 32:
+		index = 4;
+		break;
+	case 9 ... 16:
+		index = 3;
+		break;
+	case 5 ... 8:
+		index = 2;
+		break;
+	case 1 ... 4:
+		index = 1;
+	case 0:
+		break;
+	}
+
+	map[index]++;
+}
+
+void io_u_mark_submit(struct thread_data *td, unsigned int nr)
+{
+	__io_u_mark_map(td->ts.io_u_submit, nr);
+	td->ts.total_submit++;
+}
+
+void io_u_mark_complete(struct thread_data *td, unsigned int nr)
+{
+	__io_u_mark_map(td->ts.io_u_complete, nr);
+	td->ts.total_complete++;
+}
+
+void io_u_mark_depth(struct thread_data *td, unsigned int nr)
+{
+	int index = 0;
 
 	switch (td->cur_depth) {
 	default:
@@ -446,8 +518,7 @@ void io_u_mark_depth(struct thread_data *td, struct io_u *io_u)
 		break;
 	}
 
-	td->ts.io_u_map[index]++;
-	td->ts.total_io_u[io_u->ddir]++;
+	td->ts.io_u_map[index] += nr;
 }
 
 static void io_u_mark_lat_usec(struct thread_data *td, unsigned long usec)
@@ -556,8 +627,9 @@ static struct fio_file *get_next_file_rand(struct thread_data *td, int goodf,
 	do {
 		long r = os_random_long(&td->next_file_state);
 
-		fno = (unsigned int) ((double) td->o.nr_files * (r / (RAND_MAX + 1.0)));
-		f = &td->files[fno];
+		fno = (unsigned int) ((double) td->o.nr_files
+			* (r / (OS_RAND_MAX + 1.0)));
+		f = td->files[fno];
 		if (f->flags & FIO_FILE_DONE)
 			continue;
 
@@ -578,7 +650,7 @@ static struct fio_file *get_next_file_rr(struct thread_data *td, int goodf,
 	struct fio_file *f;
 
 	do {
-		f = &td->files[td->next_file];
+		f = td->files[td->next_file];
 
 		td->next_file++;
 		if (td->next_file >= td->o.nr_files)
@@ -606,7 +678,10 @@ static struct fio_file *get_next_file(struct thread_data *td)
 	assert(td->o.nr_files <= td->files_index);
 
 	if (!td->nr_open_files || td->nr_done_files >= td->o.nr_files) {
-		dprint(FD_FILE, "get_next_file: nr_open=%d, nr_done=%d, nr_files=%d\n", td->nr_open_files, td->nr_done_files, td->o.nr_files);
+		dprint(FD_FILE, "get_next_file: nr_open=%d, nr_done=%d,"
+				" nr_files=%d\n", td->nr_open_files,
+						  td->nr_done_files,
+						  td->o.nr_files);
 		return NULL;
 	}
 
@@ -657,6 +732,16 @@ set_file:
 		if (!fill_io_u(td, io_u))
 			break;
 
+		/*
+		 * optimization to prevent close/open of the same file. This
+		 * way we preserve queueing etc.
+		 */
+		if (td->o.nr_files == 1 && td->o.time_based) {
+			put_file_log(td, f);
+			fio_file_reset(f);
+			goto set_file;
+		}
+
 		/*
 		 * td_io_close() does a put_file() as well, so no need to
 		 * do that here.
@@ -689,10 +774,10 @@ struct io_u *__get_io_u(struct thread_data *td)
 {
 	struct io_u *io_u = NULL;
 
-	if (!list_empty(&td->io_u_requeues))
-		io_u = list_entry(td->io_u_requeues.next, struct io_u, list);
+	if (!flist_empty(&td->io_u_requeues))
+		io_u = flist_entry(td->io_u_requeues.next, struct io_u, list);
 	else if (!queue_full(td)) {
-		io_u = list_entry(td->io_u_freelist.next, struct io_u, list);
+		io_u = flist_entry(td->io_u_freelist.next, struct io_u, list);
 
 		io_u->buflen = 0;
 		io_u->resid = 0;
@@ -705,8 +790,8 @@ struct io_u *__get_io_u(struct thread_data *td)
 		io_u->flags &= ~IO_U_F_FREE;
 
 		io_u->error = 0;
-		list_del(&io_u->list);
-		list_add(&io_u->list, &td->io_u_busylist);
+		flist_del(&io_u->list);
+		flist_add(&io_u->list, &td->io_u_busylist);
 		td->cur_depth++;
 	}
 
@@ -744,7 +829,7 @@ struct io_u *get_io_u(struct thread_data *td)
 		dprint(FD_IO, "io_u %p, setting file failed\n", io_u);
 		goto err_put;
 	}
-	
+
 	f = io_u->file;
 	assert(f->flags & FIO_FILE_OPEN);
 
@@ -758,16 +843,18 @@ struct io_u *get_io_u(struct thread_data *td)
 
 		if (td->o.verify != VERIFY_NONE)
 			populate_verify_io_u(td, io_u);
+		else if (td->o.refill_buffers && io_u->ddir == DDIR_WRITE)
+			io_u_fill_buffer(td, io_u, io_u->xfer_buflen);
 	}
 
 	/*
 	 * Set io data pointers.
 	 */
 	io_u->endpos = io_u->offset + io_u->buflen;
-out:
 	io_u->xfer_buf = io_u->buf;
 	io_u->xfer_buflen = io_u->buflen;
 
+out:
 	if (!td_io_prep(td, io_u)) {
 		fio_gettime(&io_u->start_time, NULL);
 		return io_u;
@@ -789,7 +876,8 @@ void io_u_log_error(struct thread_data *td, struct io_u *io_u)
 
 	log_err(": %s\n", strerror(io_u->error));
 
-	log_err("     %s offset=%llu, buflen=%lu\n", msg[io_u->ddir], io_u->offset, io_u->xfer_buflen);
+	log_err("     %s offset=%llu, buflen=%lu\n", msg[io_u->ddir],
+					io_u->offset, io_u->xfer_buflen);
 
 	if (!td->error)
 		td_verror(td, io_u->error, "io_u error");
@@ -821,13 +909,13 @@ static void io_completed(struct thread_data *td, struct io_u *io_u,
 		td->io_bytes[idx] += bytes;
 		td->this_io_bytes[idx] += bytes;
 
-		io_u->file->last_completed_pos = io_u->endpos;
+		if (ramp_time_over(td)) {
+			usec = utime_since(&io_u->issue_time, &icd->time);
 
-		usec = utime_since(&io_u->issue_time, &icd->time);
-
-		add_clat_sample(td, idx, usec);
-		add_bw_sample(td, idx, &icd->time);
-		io_u_mark_latency(td, usec);
+			add_clat_sample(td, idx, usec);
+			add_bw_sample(td, idx, &icd->time);
+			io_u_mark_latency(td, usec);
+		}
 
 		if (td_write(td) && idx == DDIR_WRITE &&
 		    td->o.do_verify &&
@@ -892,19 +980,19 @@ long io_u_sync_complete(struct thread_data *td, struct io_u *io_u)
 /*
  * Called to complete min_events number of io for the async engines.
  */
-long io_u_queued_complete(struct thread_data *td, int min_events)
+long io_u_queued_complete(struct thread_data *td, int min_evts)
 {
 	struct io_completion_data icd;
 	struct timespec *tvp = NULL;
 	int ret;
 	struct timespec ts = { .tv_sec = 0, .tv_nsec = 0, };
 
-	dprint(FD_IO, "io_u_queued_completed: min=%d\n", min_events);
+	dprint(FD_IO, "io_u_queued_completed: min=%d\n", min_evts);
 
-	if (!min_events)
+	if (!min_evts)
 		tvp = &ts;
 
-	ret = td_io_getevents(td, min_events, td->cur_depth, tvp);
+	ret = td_io_getevents(td, min_evts, td->o.iodepth_batch_complete, tvp);
 	if (ret < 0) {
 		td_verror(td, -ret, "td_io_getevents");
 		return ret;
@@ -931,6 +1019,23 @@ void io_u_queued(struct thread_data *td, struct io_u *io_u)
 	add_slat_sample(td, io_u->ddir, slat_time);
 }
 
+/*
+ * "randomly" fill the buffer contents
+ */
+void io_u_fill_buffer(struct thread_data *td, struct io_u *io_u,
+		      unsigned int max_bs)
+{
+	long *ptr = io_u->buf;
+
+	if (!td->o.zero_buffers) {
+		while ((void *) ptr - io_u->buf < max_bs) {
+			*ptr = rand() * GOLDEN_RATIO_PRIME;
+			ptr++;
+		}
+	} else
+		memset(ptr, 0, max_bs);
+}
+
 #ifdef FIO_USE_TIMEOUT
 void io_u_set_timeout(struct thread_data *td)
 {
@@ -950,7 +1055,10 @@ static void io_u_dump(struct io_u *io_u)
 	unsigned long t_issue = mtime_since_now(&io_u->issue_time);
 
 	log_err("io_u=%p, t_start=%lu, t_issue=%lu\n", io_u, t_start, t_issue);
-	log_err("  buf=%p/%p, len=%lu/%lu, offset=%llu\n", io_u->buf, io_u->xfer_buf, io_u->buflen, io_u->xfer_buflen, io_u->offset);
+	log_err("  buf=%p/%p, len=%lu/%lu, offset=%llu\n", io_u->buf,
+						io_u->xfer_buf, io_u->buflen,
+						io_u->xfer_buflen,
+						io_u->offset);
 	log_err("  ddir=%d, fname=%s\n", io_u->ddir, io_u->file->file_name);
 }
 #else
@@ -964,7 +1072,7 @@ static void io_u_timeout_handler(int fio_unused sig)
 {
 	struct thread_data *td, *__td;
 	pid_t pid = getpid();
-	struct list_head *entry;
+	struct flist_head *entry;
 	struct io_u *io_u;
 	int i;
 
@@ -993,8 +1101,8 @@ static void io_u_timeout_handler(int fio_unused sig)
 
 	log_err("fio: io_u timeout: job=%s, pid=%d\n", td->o.name, td->pid);
 
-	list_for_each(entry, &td->io_u_busylist) {
-		io_u = list_entry(entry, struct io_u, list);
+	flist_for_each(entry, &td->io_u_busylist) {
+		io_u = flist_entry(entry, struct io_u, list);
 
 		io_u_dump(io_u);
 	}
