@@ -11,6 +11,7 @@
 #include "crc/md5.h"
 #include "crc/crc64.h"
 #include "crc/crc32.h"
+#include "crc/crc32c.h"
 #include "crc/crc16.h"
 #include "crc/crc7.h"
 #include "crc/sha256.h"
@@ -112,7 +113,7 @@ static void hexdump(void *buffer, int len)
  */
 static inline unsigned int __hdr_size(int verify_type)
 {
-	unsigned int len;
+	unsigned int len = len;
 
 	switch (verify_type) {
 	case VERIFY_NONE:
@@ -125,7 +126,9 @@ static inline unsigned int __hdr_size(int verify_type)
 	case VERIFY_CRC64:
 		len = sizeof(struct vhdr_crc64);
 		break;
+	case VERIFY_CRC32C:
 	case VERIFY_CRC32:
+	case VERIFY_CRC32C_INTEL:
 		len = sizeof(struct vhdr_crc32);
 		break;
 	case VERIFY_CRC16:
@@ -167,14 +170,13 @@ static void *hdr_priv(struct verify_header *hdr)
  * Return data area 'header_num'
  */
 static inline void *io_u_verify_off(struct verify_header *hdr,
-				    struct io_u *io_u,
-				    unsigned char header_num)
+				    struct io_u *io_u, unsigned char header_num)
 {
 	return io_u->buf + header_num * hdr->len + hdr_size(hdr);
 }
 
 static int verify_io_u_meta(struct verify_header *hdr, struct thread_data *td,
-                            struct io_u *io_u, unsigned int header_num)
+			    struct io_u *io_u, unsigned int header_num)
 {
 	struct vhdr_meta *vh = hdr_priv(hdr);
 
@@ -182,8 +184,7 @@ static int verify_io_u_meta(struct verify_header *hdr, struct thread_data *td,
 
 	if (vh->offset != io_u->offset + header_num * td->o.verify_interval) {
 		log_err("meta: verify failed at %llu/%u\n",
-		              io_u->offset + header_num * hdr->len,
-		              hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		return EIO;
 	}
 
@@ -207,8 +208,7 @@ static int verify_io_u_sha512(struct verify_header *hdr, struct io_u *io_u,
 
 	if (memcmp(vh->sha512, sha512_ctx.buf, sizeof(sha512))) {
 		log_err("sha512: verify failed at %llu/%u\n",
-		              io_u->offset + header_num * hdr->len,
-		              hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		hexdump(vh->sha512, sizeof(vh->sha512));
 		hexdump(sha512_ctx.buf, sizeof(sha512));
 		return EIO;
@@ -234,8 +234,7 @@ static int verify_io_u_sha256(struct verify_header *hdr, struct io_u *io_u,
 
 	if (memcmp(vh->sha256, sha256_ctx.buf, sizeof(sha256))) {
 		log_err("sha256: verify failed at %llu/%u\n",
-		              io_u->offset + header_num * hdr->len,
-		              hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		hexdump(vh->sha256, sizeof(vh->sha256));
 		hexdump(sha256_ctx.buf, sizeof(sha256));
 		return EIO;
@@ -245,7 +244,7 @@ static int verify_io_u_sha256(struct verify_header *hdr, struct io_u *io_u,
 }
 
 static int verify_io_u_crc7(struct verify_header *hdr, struct io_u *io_u,
-                            unsigned char header_num)
+			    unsigned char header_num)
 {
 	void *p = io_u_verify_off(hdr, io_u, header_num);
 	struct vhdr_crc7 *vh = hdr_priv(hdr);
@@ -257,8 +256,7 @@ static int verify_io_u_crc7(struct verify_header *hdr, struct io_u *io_u,
 
 	if (c != vh->crc7) {
 		log_err("crc7: verify failed at %llu/%u\n",
-		                io_u->offset + header_num * hdr->len,
-				hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		log_err("crc7: wanted %x, got %x\n", vh->crc7, c);
 		return EIO;
 	}
@@ -267,7 +265,7 @@ static int verify_io_u_crc7(struct verify_header *hdr, struct io_u *io_u,
 }
 
 static int verify_io_u_crc16(struct verify_header *hdr, struct io_u *io_u,
-                             unsigned int header_num)
+			     unsigned int header_num)
 {
 	void *p = io_u_verify_off(hdr, io_u, header_num);
 	struct vhdr_crc16 *vh = hdr_priv(hdr);
@@ -279,8 +277,7 @@ static int verify_io_u_crc16(struct verify_header *hdr, struct io_u *io_u,
 
 	if (c != vh->crc16) {
 		log_err("crc16: verify failed at %llu/%u\n",
-		                io_u->offset + header_num * hdr->len,
-				hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		log_err("crc16: wanted %x, got %x\n", vh->crc16, c);
 		return EIO;
 	}
@@ -289,7 +286,7 @@ static int verify_io_u_crc16(struct verify_header *hdr, struct io_u *io_u,
 }
 
 static int verify_io_u_crc64(struct verify_header *hdr, struct io_u *io_u,
-                             unsigned int header_num)
+			     unsigned int header_num)
 {
 	void *p = io_u_verify_off(hdr, io_u, header_num);
 	struct vhdr_crc64 *vh = hdr_priv(hdr);
@@ -303,7 +300,8 @@ static int verify_io_u_crc64(struct verify_header *hdr, struct io_u *io_u,
 		log_err("crc64: verify failed at %llu/%u\n",
 				io_u->offset + header_num * hdr->len,
 				hdr->len);
-		log_err("crc64: wanted %llx, got %llx\n", (unsigned long long) vh->crc64, c);
+		log_err("crc64: wanted %llx, got %llx\n",
+					(unsigned long long) vh->crc64, c);
 		return EIO;
 	}
 
@@ -311,7 +309,7 @@ static int verify_io_u_crc64(struct verify_header *hdr, struct io_u *io_u,
 }
 
 static int verify_io_u_crc32(struct verify_header *hdr, struct io_u *io_u,
-		             unsigned int header_num)
+			     unsigned int header_num)
 {
 	void *p = io_u_verify_off(hdr, io_u, header_num);
 	struct vhdr_crc32 *vh = hdr_priv(hdr);
@@ -323,8 +321,7 @@ static int verify_io_u_crc32(struct verify_header *hdr, struct io_u *io_u,
 
 	if (c != vh->crc32) {
 		log_err("crc32: verify failed at %llu/%u\n",
-		                io_u->offset + header_num * hdr->len,
-				hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		log_err("crc32: wanted %x, got %x\n", vh->crc32, c);
 		return EIO;
 	}
@@ -332,8 +329,32 @@ static int verify_io_u_crc32(struct verify_header *hdr, struct io_u *io_u,
 	return 0;
 }
 
+static int verify_io_u_crc32c(struct verify_header *hdr, struct io_u *io_u,
+			      unsigned int header_num)
+{
+	void *p = io_u_verify_off(hdr, io_u, header_num);
+	struct vhdr_crc32 *vh = hdr_priv(hdr);
+	uint32_t c;
+
+	dprint(FD_VERIFY, "crc32c verify io_u %p, len %u\n", io_u, hdr->len);
+
+	if (hdr->verify_type == VERIFY_CRC32C_INTEL)
+		c = crc32c_intel(p, hdr->len - hdr_size(hdr));
+	else
+		c = crc32c(p, hdr->len - hdr_size(hdr));
+
+	if (c != vh->crc32) {
+		log_err("crc32c: verify failed at %llu/%u\n",
+				io_u->offset + header_num * hdr->len, hdr->len);
+		log_err("crc32c: wanted %x, got %x\n", vh->crc32, c);
+		return EIO;
+	}
+
+	return 0;
+}
+
 static int verify_io_u_md5(struct verify_header *hdr, struct io_u *io_u,
-		           unsigned int header_num)
+			   unsigned int header_num)
 {
 	void *p = io_u_verify_off(hdr, io_u, header_num);
 	struct vhdr_md5 *vh = hdr_priv(hdr);
@@ -349,8 +370,7 @@ static int verify_io_u_md5(struct verify_header *hdr, struct io_u *io_u,
 
 	if (memcmp(vh->md5_digest, md5_ctx.hash, sizeof(hash))) {
 		log_err("md5: verify failed at %llu/%u\n",
-		              io_u->offset + header_num * hdr->len,
-		              hdr->len);
+				io_u->offset + header_num * hdr->len, hdr->len);
 		hexdump(vh->md5_digest, sizeof(vh->md5_digest));
 		hexdump(md5_ctx.hash, sizeof(hash));
 		return EIO;
@@ -368,7 +388,7 @@ static unsigned int hweight8(unsigned int w)
 }
 
 int verify_io_u_pattern(unsigned long pattern, unsigned long pattern_size,
-                        char *buf, unsigned int len, unsigned int mod)
+			char *buf, unsigned int len, unsigned int mod)
 {
 	unsigned int i;
 	char split_pattern[4];
@@ -411,7 +431,8 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 		hdr_inc = td->o.verify_interval;
 
 	ret = 0;
-	for (p = io_u->buf; p < io_u->buf + io_u->buflen; p += hdr_inc, hdr_num++) {
+	for (p = io_u->buf; p < io_u->buf + io_u->buflen;
+	     p += hdr_inc, hdr_num++) {
 		if (ret && td->o.verify_fatal) {
 			td->terminate = 1;
 			break;
@@ -430,10 +451,10 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 			dprint(FD_VERIFY, "pattern verify io_u %p, len %u\n",
 								io_u, hdr->len);
 			ret = verify_io_u_pattern(td->o.verify_pattern,
-			                          td->o.verify_pattern_bytes,
-			                          p + hdr_size,
-			                          hdr_inc - hdr_size,
-			                          hdr_size % 4);
+						  td->o.verify_pattern_bytes,
+						  p + hdr_size,
+						  hdr_inc - hdr_size,
+						  hdr_size % 4);
 			if (ret)
 				log_err("fio: verify failed at %llu/%u\n",
 					io_u->offset + hdr_num * hdr->len,
@@ -448,6 +469,10 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 		case VERIFY_CRC64:
 			ret = verify_io_u_crc64(hdr, io_u, hdr_num);
 			break;
+		case VERIFY_CRC32C:
+		case VERIFY_CRC32C_INTEL:
+			ret = verify_io_u_crc32c(hdr, io_u, hdr_num);
+			break;
 		case VERIFY_CRC32:
 			ret = verify_io_u_crc32(hdr, io_u, hdr_num);
 			break;
@@ -476,7 +501,7 @@ int verify_io_u(struct thread_data *td, struct io_u *io_u)
 }
 
 static void fill_meta(struct verify_header *hdr, struct thread_data *td,
-                      struct io_u *io_u, unsigned int header_num)
+		      struct io_u *io_u, unsigned int header_num)
 {
 	struct vhdr_meta *vh = hdr_priv(hdr);
 
@@ -533,6 +558,16 @@ static void fill_crc32(struct verify_header *hdr, void *p, unsigned int len)
 	vh->crc32 = crc32(p, len);
 }
 
+static void fill_crc32c(struct verify_header *hdr, void *p, unsigned int len)
+{
+	struct vhdr_crc32 *vh = hdr_priv(hdr);
+
+	if (hdr->verify_type == VERIFY_CRC32C_INTEL)
+		vh->crc32 = crc32c_intel(p, len);
+	else
+		vh->crc32 = crc32c(p, len);
+}
+
 static void fill_crc64(struct verify_header *hdr, void *p, unsigned int len)
 {
 	struct vhdr_crc64 *vh = hdr_priv(hdr);
@@ -570,7 +605,7 @@ void populate_verify_io_u(struct thread_data *td, struct io_u *io_u)
 	if (td->o.verify_interval)
 		hdr_inc = td->o.verify_interval;
 
-	for (;p < io_u->buf + io_u->buflen; p += hdr_inc) {
+	for (; p < io_u->buf + io_u->buflen; p += hdr_inc) {
 		hdr = p;
 
 		hdr->fio_magic = FIO_HDR_MAGIC;
@@ -590,6 +625,12 @@ void populate_verify_io_u(struct thread_data *td, struct io_u *io_u)
 							io_u, hdr->len);
 			fill_crc64(hdr, data, data_len);
 			break;
+		case VERIFY_CRC32C:
+		case VERIFY_CRC32C_INTEL:
+			dprint(FD_VERIFY, "fill crc32c io_u %p, len %u\n",
+							io_u, hdr->len);
+			fill_crc32c(hdr, data, data_len);
+			break;
 		case VERIFY_CRC32:
 			dprint(FD_VERIFY, "fill crc32 io_u %p, len %u\n",
 							io_u, hdr->len);
@@ -645,9 +686,9 @@ int get_next_verify(struct thread_data *td, struct io_u *io_u)
 
 		ipo = rb_entry(n, struct io_piece, rb_node);
 		rb_erase(n, &td->io_hist_tree);
-	} else if (!list_empty(&td->io_hist_list)) {
-		ipo = list_entry(td->io_hist_list.next, struct io_piece, list);
-		list_del(&ipo->list);
+	} else if (!flist_empty(&td->io_hist_list)) {
+		ipo = flist_entry(td->io_hist_list.next, struct io_piece, list);
+		flist_del(&ipo->list);
 	}
 
 	if (ipo) {
